"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family data container class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- contains all instances of the following storage classes from reports read from files:

    - FamilyBaseDataStorage
    - CategoryDataStorage
    - WarningsDataStorage
    - LinePatternDataStorage
    - SharedParameterDataStorage

- These storage instances belong to a specific family only identified by unique combination of family name nesting path and family category nesting path
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import base

import System
import json
from duHast.Revit.Family.Data.Objects.family_base_data_storage import (
    FamilyBaseDataStorage,
)
from duHast.Revit.Categories.Data.Objects.category_data_storage import (
    FamilyCategoryDataStorage,
)
from duHast.Revit.LinePattern.Data.Objects.line_pattern_data_storage import (
    FamilyLinePatternDataStorage,
)
from duHast.Revit.SharedParameters.Data.Objects.shared_parameter_data_storage import (
    FamilySharedParameterDataStorage,
)
from duHast.Revit.Warnings.Data.Objects.warnings_data_storage import (
    FamilyWarningsDataStorage,
)
from duHast.Revit.Family.Data.Objects.ifamily_data_storage import IFamilyDataStorage
from duHast.Revit.Family.Data.Objects.family_base_data_processor_defaults import (
    NESTING_SEPARATOR,
)


class FamilyDataContainer(base.Base):
    def __init__(
        self,
        family_base_data_storage=None,
        category_data_storage=None,
        line_pattern_data_storage=None,
        shared_parameter_data_storage=None,
        warnings_data_storage=None,
        **kwargs
    ):
        """
        Family data container class.

        Contains all instances of the following storage classes from reports read from files:

                - FamilyBaseDataStorage
                - CategoryDataStorage
                - WarningsDataStorage
                - LinePatternDataStorage
                - SharedParameterDataStorage

        Note: There is the possibility that only one of the storage classes is populated, this is due to the fact that the data is read from separate files and not all data is always present in each file.

        :param family_base_data_storage: a list of FamilyBaseDataStorage instances
        :type family_base_data_storage: [:class:`.FamilyBaseDataStorage`]
        :param category_data_storage: a list of FamilyCategoryDataStorage instances
        :type category_data_storage: [:class:`.FamilyCategoryDataStorage`]
        :param line_pattern_data_storage: a list of FamilyLinePatternDataStorage instances
        :type line_pattern_data_storage: [:class:`.FamilyLinePatternDataStorage`]
        :param shared_parameter_data_storage: a list of FamilySharedParameterDataStorage instances
        :type shared_parameter_data_storage: [:class:`.FamilySharedParameterDataStorage`]
        :param warnings_data_storage: a list of FamilyWarningsDataStorage instances
        :type warnings_data_storage: [:class:`.FamilyWarningsDataStorage`]
        :param kwargs: _description_
        :type kwargs: _type_
        :raises ValueError: type mismatch if any of the past in lists are not of the correct type
        """

        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!
        super(FamilyDataContainer, self).__init__(**kwargs)

        # set some base properties to their default values
        self.family_name = None
        self.family_nesting_path = None
        self.family_category = None
        self.family_category_nesting_path = None
        self.is_root_family = False
        self.family_file_path = None

        # populate class from past in data
        # start with family base data storage
        # if a valid family base data storage instance is passed in this will also set the following properties:
        # family_name, family_nesting_path, family_category, family_category_nesting_path, is_root_family, family_file_path

        # family base data default is an empty list
        self.family_base_data_storage = []
        if family_base_data_storage is None:
            # no change required
            pass
        else:
            if isinstance(family_base_data_storage, list):
                # should only ever be one entry!
                if len(family_base_data_storage) > 1:
                    raise ValueError(
                        "family_base_data_storage can only ever have one entry"
                    )
                # add to class property
                for storage_instance in family_base_data_storage:
                    self.add_family_base_data_storage(storage_instance)
            elif isinstance(family_base_data_storage, FamilyBaseDataStorage):
                # add to class property
                self.add_family_base_data_storage(family_base_data_storage)
            else:
                raise ValueError(
                    "family_base_data_storage must be a list or FamilyBaseDataStorage"
                )

        # category data storage default is an empty list
        self.category_data_storage = []
        if category_data_storage is None:
            # no change required
            pass
        else:
            if isinstance(category_data_storage, list):
                # add to class property
                for storage_instance in category_data_storage:
                    self.add_category_data_storage(storage_instance)
            else:
                self.add_category_data_storage(category_data_storage)

        # line pattern data storage default is an empty list
        self.line_pattern_data_storage = []
        if line_pattern_data_storage is None:
            # no change required
            pass
        elif isinstance(line_pattern_data_storage, list):
            # add to class property
            for storage_instance in line_pattern_data_storage:
                self.add_line_pattern_data_storage(storage_instance)
        else:
            self.add_line_pattern_data_storage(line_pattern_data_storage)

        # shared parameter data storage default is an empty list
        self.shared_parameter_data_storage = []
        if shared_parameter_data_storage is None:
            # no change required
            pass
        elif isinstance(shared_parameter_data_storage, list):
            # add to class property
            for storage_instance in shared_parameter_data_storage:
                self.add_shared_parameter_data_storage(storage_instance)
        else:
            self.add_shared_parameter_data_storage(shared_parameter_data_storage)

        # warnings data storage default is an empty list
        self.warnings_data_storage = []
        if warnings_data_storage is None:
            # no change required
            pass
        elif isinstance(warnings_data_storage, list):
            # add to class property
            for storage_instance in warnings_data_storage:
                self.add_warnings_data_storage(storage_instance)
        else:
            self.add_warnings_data_storage(warnings_data_storage)

    def _update_base_properties_from_storage(self, storage_instance):
        """
        Will update the base properties of the class from a family base data storage instance.

        :param storage_instance: a family data storage instance
        :type storage_instance: :class:`.IFamilyDataStorage`
        """

        # set other class properties based on storage
        self.family_name = storage_instance.family_name
        self.family_file_path = storage_instance.family_file_path
        self.family_nesting_path = storage_instance.root_name_path
        self.family_category_nesting_path = storage_instance.root_category_path

        # set the category property based on the last entry in root category path
        category_chunks = storage_instance.root_category_path.split(NESTING_SEPARATOR)
        self.family_category = category_chunks[-1]

        # check if this is a root family
        if NESTING_SEPARATOR in storage_instance.root_name_path:
            self.is_root_family = False
        else:
            self.is_root_family = True

    def add_family_base_data_storage(self, other):
        """
        Will add a family base data storage instance and infer some other class properties from it.

        Note: if storage data properties root_name_path and root_category_path are different to the current values
        all other storage properties will be wiped to avoid mismatches!

        :param other: A family base data storage instance added to this class
        :type other: :class:`.FamilyBaseDataStorage`
        :raises TypeError: If other is not an instance of FamilyBaseDataStorage a type error will be raised.
        """
        # check is correct object type
        if isinstance(other, FamilyBaseDataStorage) == False:
            raise TypeError("other must be a list of FamilyBaseDataStorage")

        # only add if not already in list
        if other not in self.family_base_data_storage:
            # there is always only going to be one entry in list, pop the existing one if there is one
            if len(self.family_base_data_storage) > 0:
                self.family_base_data_storage.pop()

            # add new one to class property
            self.family_base_data_storage.append(other)

            # check if nesting path and category nesting path are different to the current values but not None!
            # if so wipe the other storage properties to avoid mismatches!
            if (
                self.family_nesting_path != other.root_name_path
                and self.family_nesting_path != None
                or self.family_category_nesting_path != other.root_category_path
                and self.family_category_nesting_path != None
            ):
                self.category_data_storage = []
                self.line_pattern_data_storage = []
                self.shared_parameter_data_storage = []
                self.warnings_data_storage = []

            # set other class properties based on storage
            self._update_base_properties_from_storage(other)
        else:
            raise ValueError(
                "Base storage is not unique: family nesting path: {} category nesting path: {}".format(
                    other.root_name_path, other.root_category_path
                )
            )

    def add_category_data_storage(self, other):
        """
        Will add a category data storage instance.

        Note: an exception will be raised if family_nesting_path and family_category_nesting_path do not match the current values!

        :param other: a data storage instance
        :type other: :class:`.FamilyCategoryDataStorage`
        :raises TypeError: If other is not an instance of FamilyCategoryDataStorage a type error will be raised.
        """
        # check is correct object type
        if isinstance(other, FamilyCategoryDataStorage) == False:
            raise TypeError("other must be a list of FamilyCategoryDataStorage")

        # only add if not already in list
        if other not in self.category_data_storage:
            # check if nesting path and category nesting path are different to the current values but not None!
            # if so  throw error!
            if (
                self.family_nesting_path != other.root_name_path
                and self.family_nesting_path != None
                or self.family_category_nesting_path != other.root_category_path
                and self.family_category_nesting_path != None
            ):
                raise ValueError(
                    "other root_name_path and root_category_path must match current values"
                )
            elif (
                self.family_nesting_path == None
                and self.family_category_nesting_path == None
            ):
                # looks like this might be the only storage class added or
                # family base data storage is absent and this is the first storage class added
                self._update_base_properties_from_storage(other)

            # add to class property only if unique (unique is if category name, sub category name and sub category id are unique)
            if len(self.category_data_storage) == 0:
                # add to class property
                self.category_data_storage.append(other)
            else:
                found_match = False
                for entry in self.category_data_storage:
                    if (
                        entry.category_name == other.category_name
                        and entry.sub_category_name == other.sub_category_name
                        and entry.sub_category_id == other.sub_category_id
                    ):
                        found_match = True
                        break
                if not found_match:
                    # add to class property
                    self.category_data_storage.append(other)
                else:
                    raise ValueError(
                        "Category storage is not unique: category: {} sub category: {} id: {}".format(
                            other.category_name,
                            other.sub_category_name,
                            other.sub_category_id,
                        )
                    )
        else:
            raise ValueError(
                "Category storage is not unique: category: {} sub category: {} id: {}".format(
                    other.category_name, other.sub_category_name, other.sub_category_id
                )
            )

    def add_line_pattern_data_storage(self, other):
        """
        Will add a line pattern data storage instance.

        Note: an exception will be raised if family_nesting_path and family_category_nesting_path do not match the current values!

        :param other: a data storage instance
        :type other: :class:`.FamilyLinePatternDataStorage`
        :raises TypeError: If other is not an instance of FamilyLinePatternDataStorage a type error will be raised.
        """
        # check is correct object type
        if isinstance(other, FamilyLinePatternDataStorage) == False:
            raise TypeError("other must be a list of FamilyLinePatternDataStorage")

        # only add if not already in list
        if other not in self.line_pattern_data_storage:
            # check if nesting path and category nesting path are different to the current values but not None!
            # if so throw error!
            if (
                self.family_nesting_path != other.root_name_path
                and self.family_nesting_path != None
                or self.family_category_nesting_path != other.root_category_path
                and self.family_category_nesting_path != None
            ):
                raise ValueError(
                    "other root_name_path and root_category_path must match current values"
                )
            elif (
                self.family_nesting_path == None
                and self.family_category_nesting_path == None
            ):
                # looks like this might be the only storage class added or
                # family base data storage is absent and this is the first storage class added
                self._update_base_properties_from_storage(other)

            # add to class property only if unique (unique is if pattern name and pattern id are unique)
            if len(self.line_pattern_data_storage) == 0:
                # add to class property
                self.line_pattern_data_storage.append(other)
            else:
                found_match = False
                for entry in self.line_pattern_data_storage:
                    if (
                        entry.pattern_name == other.pattern_name
                        and entry.pattern_id == other.pattern_id
                    ):
                        found_match = True
                        break
                if not found_match:
                    # add to class property
                    self.line_pattern_data_storage.append(other)
                else:
                    raise ValueError(
                        "Line pattern storage is not unique: name: {} id: {}".format(
                            other.pattern_name, other.pattern_id
                        )
                    )
        else:
            raise ValueError(
                "Line pattern storage is not unique: name: {} id: {}".format(
                    other.pattern_name, other.pattern_id
                )
            )

    def add_shared_parameter_data_storage(self, other):
        """
        Will add a shared parameter data storage instance.

        Note: an exception will be raised if family_nesting_path and family_category_nesting_path do not match the current values!

        :param other: a data storage instance
        :type other: :class:`.FamilySharedParameterDataStorage`
        :raises TypeError: If other is not an instance of FamilySharedParameterDataStorage a type error will be raised.
        """
        # check is correct object type
        if isinstance(other, FamilySharedParameterDataStorage) == False:
            raise TypeError("other must be a list of FamilySharedParameterDataStorage")

        # only add if not already in list
        if other not in self.shared_parameter_data_storage:
            # check if nesting path and category nesting path are different to the current values but not None!
            # if so wipe throw error!
            if (
                self.family_nesting_path != other.root_name_path
                and self.family_nesting_path != None
                or self.family_category_nesting_path != other.root_category_path
                and self.family_category_nesting_path != None
            ):
                raise ValueError(
                    "other root_name_path and root_category_path must match current values"
                )
            elif (
                self.family_nesting_path == None
                and self.family_category_nesting_path == None
            ):
                # looks like this might be the only storage class added or
                # family base data storage is absent and this is the first storage class added
                self._update_base_properties_from_storage(other)

            # add to class property only if unique (unique is if parameter name and parameter id are unique)
            if len(self.shared_parameter_data_storage) == 0:
                # add to class property
                self.shared_parameter_data_storage.append(other)
            else:
                found_match = False
                for entry in self.shared_parameter_data_storage:
                    if (
                        entry.parameter_name == other.parameter_name
                        and entry.parameter_id == other.parameter_id
                    ):
                        found_match = True
                        break
                if not found_match:
                    # add to class property
                    self.shared_parameter_data_storage.append(other)
                else:
                    raise ValueError(
                        "Shared parameter storage is not unique: name: {} id: {}".format(
                            other.parameter_name, other.parameter_id
                        )
                    )
        else:
            raise ValueError(
                "Shared parameter storage is not unique: name: {} id: {}".format(
                    other.parameter_name, other.parameter_id
                )
            )

    def add_warnings_data_storage(self, other):
        """
        Will add a warnings data storage instance.

        Note: an exception will be raised if family_nesting_path and family_category_nesting_path do not match the current values!

        :param other: a data storage instance
        :type other: :class:`.FamilyWarningsDataStorage`
        :raises TypeError: If other is not an instance of FamilyWarningsDataStorage a type error will be raised.
        """
        # check is correct object type
        if isinstance(other, FamilyWarningsDataStorage) == False:
            raise TypeError("other must be type of FamilyWarningsDataStorage")

        # only add if not already in list
        if other not in self.warnings_data_storage:
            # check if nesting path and category nesting path are different to the current values but not None!
            # if so wipe throw error!
            if (
                self.family_nesting_path != other.root_name_path
                and self.family_nesting_path != None
                or self.family_category_nesting_path != other.root_category_path
                and self.family_category_nesting_path != None
            ):
                raise ValueError(
                    "other root_name_path and root_category_path must match current values"
                )
            elif (
                self.family_nesting_path == None
                and self.family_category_nesting_path == None
            ):
                # looks like this might be the only storage class added or
                # family base data storage is absent and this is the first storage class added
                self._update_base_properties_from_storage(other)

            # add to class property only if unique (unique is if warning text and warning guid are unique)
            if len(self.warnings_data_storage) == 0:
                # add to class property
                self.warnings_data_storage.append(other)
            else:
                found_match = False
                for entry in self.warnings_data_storage:
                    if (
                        entry.warning_text == other.warning_text
                        and entry.warning_guid == other.warning_guid
                    ):
                        found_match = True
                        break
                if not found_match:
                    # add to class property
                    self.warnings_data_storage.append(other)
                else:
                    raise ValueError(
                        "Warnings storage is not unique: text: {} guid: {}".format(
                            other.warning_text, other.warning_guid
                        )
                    )
        else:
            raise ValueError(
                "Warnings storage is not unique: text: {} guid: {}".format(
                    other.warning_text, other.warning_guid
                )
            )

    def add_data_storage(self, other):
        """
        Adds a new data storage instance to this container.

        :param other: The new data storage instance.
        :type other: :class:`.IFamilyDataStorage`
        :raises TypeError: If other is not an instance of IFamilyDataStorage a type error will be raised.
        """

        # check is correct object type
        if isinstance(other, IFamilyDataStorage) == False:
            raise TypeError(
                "other must be a type of IFamilyDataStorage but is: {}".format(
                    type(other)
                )
            )

        if other.data_type == FamilyBaseDataStorage.data_type:
            self.add_family_base_data_storage(other=other)
        elif other.data_type == FamilyCategoryDataStorage.data_type:
            self.add_category_data_storage(other=other)
        elif other.data_type == FamilyLinePatternDataStorage.data_type:
            self.add_line_pattern_data_storage(other=other)
        elif other.data_type == FamilySharedParameterDataStorage.data_type:
            self.add_shared_parameter_data_storage(other=other)
        elif other.data_type == FamilyWarningsDataStorage.data_type:
            self.add_warnings_data_storage(other=other)
        else:
            raise TypeError(
                "Data storage type : {} is not supported.".format(other.data_type)
            )

    def _get_storage_of_type(storage_property):
        """
        Loops of the storage instances saved in the storage property of this class and returns their string values

        :param storage_property: a property of this class containing IFamilyDataStorage instances
        :type storage_property: :class:`.Result` IFamilyDataStorage
        :return: A nested list of strings representing each storage instance
        :rtype: [[str]]
        """
        data = []
        for storage in storage_property:
            if isinstance(storage, IFamilyDataStorage) == False:
                raise TypeError(
                    "Storage should of type IFamilyDataStorage but is [{}]".format(
                        type(storage)
                    )
                )
            data.append(storage.get_data_values_as_list_of_strings())
        return data

    def get_family_base_data_storage_as_string(self):
        """
        Returns all family base data instance storage as string

        :return: A dictionary where key is the storage data type, and value is a nested list of strings representing each storage instance
        :rtype: {str:[[str]]}
        """
        data = self._get_storage_of_type(storage_property=self.family_base_data_storage)
        return {FamilyBaseDataStorage.data_type: data}

    def get_family_category_data_storage_as_string(self):
        """
        Returns all family category data instance storage as string

        :return: A dictionary where key is the storage data type, and value is a nested list of strings representing each storage instance
        :rtype: {str:[[str]]}
        """
        data = self._get_storage_of_type(storage_property=self.category_data_storage)
        return {FamilyCategoryDataStorage.data_type: data}

    def get_line_pattern_data_storage_as_string(self):
        """
        Returns all family line pattern data instance storage as string

        :return: A dictionary where key is the storage data type, and value is a nested list of strings representing each storage instance
        :rtype: {str:[[str]]}
        """
        data = self._get_storage_of_type(
            storage_property=self.line_pattern_data_storage
        )
        return {FamilyLinePatternDataStorage.data_type: data}

    def get_shared_parameter_data_storage_as_string(self):
        """
        Returns all family shared parameter data instance storage as string

        :return: A dictionary where key is the storage data type, and value is a nested list of strings representing each storage instance
        :rtype: {str:[[str]]}
        """
        data = self._get_storage_of_type(
            storage_property=self.shared_parameter_data_storage
        )
        return {FamilySharedParameterDataStorage.data_type: data}

    def get_warnings_data_storage_as_string(self):
        """
        Returns all family warnings data instance storage as string

        :return: A dictionary where key is the storage data type, and value is a nested list of strings representing each storage instance
        :rtype: {str:[[str]]}
        """
        data = self._get_storage_of_type(storage_property=self.warnings_data_storage)
        return {FamilyWarningsDataStorage.data_type: data}

    def get_data_string_list(self):
        """
        returns a dictionary where key is the storage data type and value is a list of string representing the data storage
        """

        return_value = {}
        storage_as_strings = []

        # get all storage data as string ( represented in dictionary where key is the data type and values list of lists containing the storage data)
        storage_as_strings.append(self.get_family_base_data_storage_as_string())
        storage_as_strings.append(self.get_family_category_data_storage_as_string())
        storage_as_strings.append(self.get_line_pattern_data_storage_as_string())
        storage_as_strings.append(self.get_shared_parameter_data_storage_as_string())
        storage_as_strings.append(self.get_warnings_data_storage_as_string())

        # build single dictionary by data type:
        for d in storage_as_strings:
            return_value[d.key] = d.value
        return return_value

    def _get_storage_headers_of_type(storage_property):
        """
        Loops of the storage instances saved in the storage property names of this class.

        :param storage_property: a property of this class containing IFamilyDataStorage instances
        :type storage_property: :class:`.Result` IFamilyDataStorage
        :return: A single nested list of strings representing storage instance property names
        :rtype: [[str]]
        """
        data = []
        for storage in storage_property:
            if isinstance(storage, IFamilyDataStorage) == False:
                raise TypeError(
                    "Storage should of type IFamilyDataStorage but is [{}]".format(
                        type(storage)
                    )
                )
            data.append(storage.get_property_names())
            # we only need one entry
            break
        return data

    def get_family_base_data_storage_headers_as_string(self):
        """
        Returns all family base data instance storage property as string

        :return: A dictionary where key is the storage data type, and value is a nested list of strings representing each storage property name
        :rtype: {str:[[str]]}
        """
        data = self._get_storage_headers_of_type(
            storage_property=self.family_base_data_storage
        )
        return {FamilyBaseDataStorage.data_type: data}

    def get_family_category_data_storage_headers_as_string(self):
        """
        Returns all family category data instance storage as string

        :return: A dictionary where key is the storage data type, and value is a nested list of strings representing each storage property name
        :rtype: {str:[[str]]}
        """
        data = self._get_storage_headers_of_type(
            storage_property=self.category_data_storage
        )
        return {FamilyCategoryDataStorage.data_type: data}

    def get_line_pattern_data_storage_headers_as_string(self):
        """
        Returns all family line pattern data instance storage as string

        :return: A dictionary where key is the storage data type, and value is a nested list of strings representing each storage property name
        :rtype: {str:[[str]]}
        """
        data = self._get_storage_headers_of_type(
            storage_property=self.line_pattern_data_storage
        )
        return {FamilyLinePatternDataStorage.data_type: data}

    def get_shared_parameter_data_storage_headers_as_string(self):
        """
        Returns all family shared parameter data instance storage as string

        :return: A dictionary where key is the storage data type, and value is a nested list of strings representing each storage property name
        :rtype: {str:[[str]]}
        """
        data = self._get_storage_headers_of_type(
            storage_property=self.shared_parameter_data_storage
        )
        return {FamilySharedParameterDataStorage.data_type: data}

    def get_warnings_data_storage_headers_as_string(self):
        """
        Returns all family warnings data instance storage as string

        :return: A dictionary where key is the storage data type, and value is a nested list of strings representing each storage property name
        :rtype: {str:[[str]]}
        """
        data = self._get_storage_headers_of_type(
            storage_property=self.warnings_data_storage
        )
        return {FamilyWarningsDataStorage.data_type: data}

    def get_data_headers_list(self):
        """
        returns a dictionary where key is the storage data type and value is a nested list of list of string representing the data storage property names

        :return: Dictionary
        :rtype: {str:[[str]]}
        """
        return_value = {}
        storage_as_strings = []

        # get all storage data as string ( represented in dictionary where key is the data type and values list of lists containing the storage data)
        storage_as_strings.append(self.get_family_base_data_storage_headers_as_string())
        storage_as_strings.append(
            self.get_family_category_data_storage_headers_as_string()
        )
        storage_as_strings.append(
            self.get_line_pattern_data_storage_headers_as_string()
        )
        storage_as_strings.append(
            self.get_shared_parameter_data_storage_headers_as_string()
        )
        storage_as_strings.append(self.get_warnings_data_storage_headers_as_string())

        # build single dictionary by data type:
        for d in storage_as_strings:
            return_value[d.key] = d.value
        return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to by element in view overrides. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common.transaction import in_transaction_with_failure_handling

from Autodesk.Revit.DB import Document, OverrideGraphicSettings, Transaction, View


def overwrite_element_graphics_in_view(
    doc, view, elements, override_graphics_settings=None
):
    """
    Overwrite the graphics of elements in a view.

    Note:
    If None is passed as the override_graphics_settings parameter, the default settings will be used to remove any overrides on the elements.

    :param doc: The Revit document.
    :type doc: Document
    :param view: The view in which the elements are located.
    :type view: View
    :param elements: The elements to overwrite the graphics for.
    :type elements: list
    :param override_graphics_settings: The override settings to apply to the elements.
    :type override_graphics_settings: Autodesk.Revit.DB.OverrideGraphicSettings

    :return: A result object.
    :rtype: :class:`.Result`
    """

    # set up a default return value object
    return_value = Result()

    try:
        # do some type checking
        if (doc is None) or (view is None) or (elements is None):
            return_value.update_sep(
                False,
                "Failed to override elements in view. One or more required parameters are None.",
            )
            return return_value

        if len(elements) == 0:
            return_value.update_sep(
                False,
                "Failed to override elements in view. The elements list is empty.",
            )
            return return_value

        if not isinstance(elements, list):
            return_value.update_sep(
                False,
                "Failed to override elements in view. The elements parameter is not a list.",
            )
            return return_value

        if not isinstance(view, View):
            return_value.update_sep(
                False,
                "Failed to override elements in view. The view parameter is not a View.",
            )
            return return_value

        if not isinstance(doc, Document):
            return_value.update_sep(
                False,
                "Failed to override elements in view. The doc parameter is not a Document.",
            )
            return return_value

        if override_graphics_settings is not None and not isinstance(
            override_graphics_settings, OverrideGraphicSettings
        ):
            return_value.update_sep(
                False,
                "Failed to override elements in view. The override_graphics_settings parameter is not None and not  an OverrideGraphicSettings object.",
            )
            return return_value

        # set up a default override settings object
        override_settings = OverrideGraphicSettings()
        if override_graphics_settings is not None:
            # set up the override settings object with the provided settings
            override_settings = override_graphics_settings

        # override the graphics for each element
        for element in elements:

            def action():
                action_return_value = Result()
                try:
                    view.SetElementOverrides(element.Id, override_settings)
                    action_return_value.append_message(
                        "Successfully overrode graphics for element: {}.".format(
                            element.Id
                        )
                    )
                except Exception as e:
                    action_return_value.update_sep(
                        False,
                        "Failed to override graphics for element: {}. Error: {}".format(
                            element.Id, e
                        ),
                    )
                return action_return_value

            tranny = Transaction(doc, "Override Graphics")
            tranny_result = in_transaction_with_failure_handling(
                transaction=tranny, action=action
            )
            return_value.update(tranny_result)
    except Exception as e:
        return_value.update_sep(
            False, "Failed to override elements in view with exception: {}".format(e)
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Data storage class for Revit room properties.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json

from duHast.Data.Objects.Collectors.Properties import data_design_set_option
from duHast.Data.Objects.Collectors.Properties import data_phasing
from duHast.Data.Objects.Collectors.Properties import data_level
from duHast.Data.Objects.Collectors.Properties import data_instance_properties
from duHast.Data.Objects.Collectors.Properties import data_revit_model
from duHast.Data.Objects.Collectors import data_base
from duHast.Data.Objects.Collectors.Properties import data_element_geometry_base
from duHast.Data.Objects.Collectors.Properties.data_property_names import (
    DataPropertyNames,
)


class DataRoom(data_base.DataBase, data_element_geometry_base.DataElementGeometryBase):
    data_type = "room"

    def __init__(self, j=None):
        """
        Class constructor.

        :param j: A json formatted dictionary of this class, defaults to {}
        :type j: dict, optional
        """

        # initialise parent classes with values
        super(DataRoom, self).__init__(data_type=DataRoom.data_type, j=j)

        # initialise classes with default values
        self.associated_elements = []
        self.instance_properties = data_instance_properties.DataInstanceProperties()
        self.level = data_level.DataLevel()
        self.revit_model = data_revit_model.DataRevitModel()
        self.phasing = data_phasing.DataPhasing()
        self.design_set_and_option = data_design_set_option.DataDesignSetOption()

        json_var = None
        # check if any data was past in with constructor!
        if j is not None:
            # check type of data that came in:
            if isinstance(j, str):
                # a string
                json_var = json.loads(j)
            elif isinstance(j, dict):
                # no action required
                json_var = j.copy()
            else:
                raise TypeError(
                    "Argument j supplied must be of type string or type dictionary. Got {} instead.".format(
                        type(j)
                    )
                )

            # attempt to populate from json
            try:
                self.instance_properties = (
                    data_instance_properties.DataInstanceProperties(
                        json_var.get(
                            data_instance_properties.DataInstanceProperties.data_type,
                            None,
                        )
                    )
                )
                self.design_set_and_option = data_design_set_option.DataDesignSetOption(
                    json_var.get(
                        data_design_set_option.DataDesignSetOption.data_type, None
                    )
                )
                self.level = data_level.DataLevel(
                    json_var.get(data_level.DataLevel.data_type, None)
                )
                self.revit_model = data_revit_model.DataRevitModel(
                    json_var.get(data_revit_model.DataRevitModel.data_type, None)
                )

                self.phasing = data_phasing.DataPhasing(
                    json_var.get(data_phasing.DataPhasing.data_type, None)
                )

                # get associated elements
                associated_elements = json_var.get(
                    DataPropertyNames.ASSOCIATED_ELEMENTS,
                    self.associated_elements,
                )
                # these can be all sorts of types...
                # TODO: convert json to actual elements

            except Exception as e:
                raise type(e)(
                    "Node {} failed to initialise with: {}".format(self.data_type, e)
                )

    def __eq__(self, other):
        """
        equal compare ( ignores associated elements property)

        Args:
            other (DataRoom): another DataRoom instance

        Returns:
            bool: True if equal, otherwise False
        """
        if not isinstance(other, DataRoom):
            return NotImplemented
        return (
            self.instance_properties == other.instance_properties
            and self.level == other.level
            and self.revit_model == other.revit_model
            and self.phasing == other.phasing
            and self.design_set_and_option == other.design_set_and_option
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(
            (
                self.instance_properties,
                self.level,
                self.revit_model,
                self.phasing,
                self.design_set_and_option,
            )
        )



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit extrusion create helper functions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

These functions work within a Revit family document to create and manipulate extrusions.


"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common.Geometry.curve_loops import convert_curve_loops_to_curve_arr_array
from duHast.Revit.Common.transaction import in_transaction

from duHast.Revit.Categories.Utility.category_property_names import CATEGORY_GRAPHIC_STYLE_3D
from duHast.Revit.Categories.categories import (
    ELEMENTS_PARAS_SUB
)

from duHast.Revit.Common.parameter_set_utils import set_parameter_value

from duHast.Revit.Family.family_parameter_utils import associate_parameter_with_other_parameter_on_nested_family_instance
from duHast.Revit.Family.Geometry.utils import set_element_sub_category
from duHast.Utilities.unit_conversion import convert_mm_to_imperial_feet

from Autodesk.Revit.DB import BuiltInParameter,Transaction, SketchPlane


def create_extrusion(doc, sketch_plane, curve_loops, func, height=100, transaction_manager=in_transaction):
    """
    Create a new extrusion in the family document using the provided curve loops.
    
    
    :param doc: The family document.
    :type doc: Autodesk.Revit.DB.Document
    :param sketch_plane: The sketch plane to use for the new extrusion.
    :type sketch_plane: Autodesk.Revit.DB.SketchPlane
    :param curve_loops: The curve loops to use for the new extrusion.
    :type curve_loops: list of Autodesk.Revit.DB.CurveLoop
    :param func: A function to execute on the new extrusion. ( need to take the doc and the new extrusion as arguments and need to return a Result class instance)
    :type func: function
    :param transaction_manager: A function to manage transactions. (default is in_transaction)
    :type transaction_manager: function
    
    :return: Result class instance.

        - `result.status` (bool): True if the extrusion was created successfully, otherwise False.
        - `result.message` (str): Confirmation of successful creation of the extrusion.
        - `result.result` (list): The new extrusion element.
        
    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Generic exception message.
        - `result.result` (list): Empty.
        
    :rtype: :class:`.Result`
    """
    
    return_value = Result()
    try:
        
        # convert filled region curve loops to curve array array
        new_profile = convert_curve_loops_to_curve_arr_array(curve_loops)
        
        # create a new extrusion in the family document
        def action():
       
            action_return_value = Result()
            try:
                # Create new extrusion
                new_extrusion = doc.FamilyCreate.NewExtrusion(True, new_profile, sketch_plane, convert_mm_to_imperial_feet(height))
                
                # check if the extrusion was created successfully
                if new_extrusion is None:
                    action_return_value.update_sep(False, "Failed to create new extrusion in family")
                else:
                    action_return_value.append_message("Created new extrusion in family")
                    
                    # execute function on extrusion
                    if func is not None:
                        func_result = func(doc, new_extrusion)
                        action_return_value.update(func_result)
                
                # add the new extrusion to the result
                action_return_value.result.append(new_extrusion)
                
            except Exception as e:
                action_return_value.update_sep(False, "Failed to create new extrusion in family: {}".format(e))
            return action_return_value

        if transaction_manager :
            transaction = Transaction(doc, "Creating extrusion")
            return_value = transaction_manager(transaction,action )
        else:
            return_value = action()
            
    except Exception as e: 
        message = "Failed to create extrusion in family: {}".format(e)
        return_value.update_sep(False, message)
    return return_value


def create_extrusion_on_level (doc, level, curve_loops, func, height=100, transaction_manager=in_transaction):
    """
    Create a new extrusion in the family document on the level provided using the provided curve loops.

    :param doc: The family document.
    :type doc: Autodesk.Revit.DB.Document
    :param level: The level to use for the new extrusion.
    :type level: Autodesk.Revit.DB.Level
    :param curve_loops: The curve loops to use for the new extrusion.
    :type curve_loops: list of Autodesk.Revit.DB.CurveLoop
    :return: Result class instance.
    :param func: A function to execute on the new extrusion. ( need to take the doc and the new extrusion as arguments and need to return a Result class instance)
    :type func: function
    :param transaction_manager: A function to manage transactions. (default is in_transaction)
    :type transaction_manager: function

        - `result.status` (bool): True if the extrusion was created successfully, otherwise False.
        - `result.message` (str): Confirmation of successful creation of the extrusion.
        - `result.result` (list): The new extrusion element.
    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Generic exception message.
        - `result.result` (list): Empty.
    :rtype: :class:`.Result`
    """
    
    return_value = Result()
    try:
        
        # convert filled region curve loops to curve array array
        new_profile = convert_curve_loops_to_curve_arr_array(curve_loops)
        
        # get the reference plane of the level
        level_reference = level.GetPlaneReference()
               
        # create a new extrusion in the family document
        def action():
       
            action_return_value = Result()
            try:
                
                # create a sketch plane on the level reference
                sketch_plane = SketchPlane.Create(doc, level_reference)
                 
                # Create new extrusion
                new_extrusion = doc.FamilyCreate.NewExtrusion(True, new_profile, sketch_plane, convert_mm_to_imperial_feet(height))
                
                # check if the extrusion was created successfully
                if new_extrusion is None:
                    action_return_value.update_sep(False, "Failed to create new extrusion in family")
                else:
                    action_return_value.append_message("Created new extrusion in family")
                    
                    # execute function on extrusion
                    if func is not None:
                        func_result = func(doc, new_extrusion)
                        action_return_value.update(func_result)
                
                # add the new extrusion to the result
                action_return_value.result.append(new_extrusion)
                
            except Exception as e:
                action_return_value.update_sep(False, "Failed to create new extrusion in family: {}".format(e))
            return action_return_value

        if transaction_manager :
            transaction = Transaction(doc, "Creating extrusion")
            return_value = transaction_manager(transaction,action )
        else:
            return_value = action()
            
    except Exception as e: 
        message = "Failed to create extrusion in family: {}".format(e)
        return_value.update_sep(False, message)
    return return_value
    
    
def associate_extrusion_height_with_parameter(doc, extrusion, height_parameter_name, transaction_manager=in_transaction):
    """
    Set the height parameter of an extrusion to a parameter in the family.

    :param doc: The family document.
    :type doc: Autodesk.Revit.DB.Document
    :param extrusion: The extrusion element.
    :type extrusion: Autodesk.Revit.DB.Extrusion
    :param height_parameter_name: The name of the height parameter in the family.

    :type height_parameter_name: str
    :return: Result class instance.

        - `result.status` (bool): True if the extrusion height parameter was set successfully, otherwise False.
        - `result.message` (str): Confirmation of successful setting of the extrusion height parameter.
        - `result.result` (list): The new extrusion element.
        

    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Generic exception message.
        - `result.result` (list): Empty.
        
    :rtype: :class:`.Result`
    """

    return_value = Result()
    try:
        # get the localised name of the height parameter
        param_extrusion_height = extrusion.get_Parameter(BuiltInParameter.EXTRUSION_END_PARAM)
        param_extrusion_height_name = param_extrusion_height.Definition.Name
        
        # set the height of the new extrusion to a parameter in the family
        attach_height_parameter_result = associate_parameter_with_other_parameter_on_nested_family_instance(
            doc=doc, 
            nested_family_instance=extrusion, 
            target_parameter_name=param_extrusion_height_name, 
            source_parameter_name=height_parameter_name,
            transaction_manager=transaction_manager
        )
        return attach_height_parameter_result
    except Exception as e:
        message = "Failed to set extrusion height parameter in family: {}".format(e)
        return_value.update_sep(False, message)
    return return_value


def set_extrusion_sub_category(doc, extrusion, source_graphic_style, transaction_manager=in_transaction):
    """
    Set the subcategory of an extrusion in a family document.
    The subcategory is set to the subcategory of the source 3D graphic style.
    
    :param doc: The family document.
    :type doc: Autodesk.Revit.DB.Document
    :param extrusion: The extrusion element.
    :type extrusion: Autodesk.Revit.DB.Extrusion
    :param source_graphic_style: The graphic style of the source extrusion.
    :type source_graphic_style: Autodesk.Revit.DB.GraphicStyle
    :return: Result class instance.

        - `result.status` (bool): True if the extrusion subcategory was set successfully, otherwise False.
        - `result.message` (str): Confirmation of successful setting of the extrusion subcategory.
        - `result.result` (list): Empty.
    
    
    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Generic exception message.
        - `result.result` (list): Empty.
        
    :rtype: :class:`.Result`
    """
        
    return_value = Result()
    try:
        return_value = set_element_sub_category(
            doc=doc,
            element=extrusion,
            source_graphic_style=source_graphic_style,
            source_graphic_style_key=CATEGORY_GRAPHIC_STYLE_3D,
            transaction_manager=transaction_manager,
        )
    except Exception as e:
        message = "Failed to set sub category in family: {}".format(e)
        return_value.update_sep(False, message)
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
View crop utilities.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common.transaction import in_transaction

from Autodesk.Revit.DB import Transaction


def get_view_crop(view):
    """
    Get the crop region loops of a view.

    This method returns a representation of the boundaries of the currently active crop for the associated view.
    If the view crop has a non-rectangular shape, the method returns that shape.
    Otherwise, if the view crop has been split, the method returns the multiple rectangular boundaries visible in the crop - note that this does not reflect any offsets that may have been applied to the boundary regions.
    If the crop is not split, this returns a single rectangle representing the crop.
    All coordinates are in the coordinate frame of the view.

    :param view: The view to get the crop region from.
    :type view: Autodesk.Revit.DB.View
    :return: The crop region of the view.
    :rtype: [Autodesk.Revit.DB.CurveLoop]
    """

    # attempt to get the views shape manager
    shape_manager = view.GetCropRegionShapeManager()

    # check if the shape manager is None
    if shape_manager is None:
        return None

    # get the crop shape loops
    loops = shape_manager.GetCropShape()

    return loops


def apply_view_crop_without_transaction_wrapper(view, crop_loop):
    """
    Apply a crop region to a view.

    This method sets the crop region of the view to the specified crop loops.

    :param view: The view to apply the crop region to.
    :type view: Autodesk.Revit.DB.View
    :param crop_loops: The crop loop to apply to the view.
    :type crop_loops: Autodesk.Revit.DB.CurveLoop

    :return: The result of the operation.
    :rtype: duHast.Utilities.Objects.result.Result
    """

    return_value = Result()

    try:
        # attempt to get the views shape manager
        shape_manager = view.GetCropRegionShapeManager()

        # check if the shape manager is None
        if shape_manager is None:
            return_value.update_sep(
                False,
                "Could not get the crop region shape manager for the view: {}.".format(
                    view
                ),
            )
            return return_value
       
        # set the crop shape
        shape_manager.SetCropShape(crop_loop)
        return_value .update_sep(
            True, "Successfully updated view crop: {}".format(view)
        )
       
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to update view crop: {} with error: {}".format(view, e),
        )

    return return_value


def apply_view_crop(doc, view, crop_loop):
    """
    Apply a crop region to a view.

    This method sets the crop region of the view to the specified crop loops.

    :param doc: The document containing the view.
    :type doc: Autodesk.Revit.DB.Document
    :param view: The view to apply the crop region to.
    :type view: Autodesk.Revit.DB.View
    :param crop_loops: The crop loop to apply to the view.
    :type crop_loops: Autodesk.Revit.DB.CurveLoop

    :return: The result of the operation.
    :rtype: duHast.Utilities.Objects.result.Result
    """

    return_value = Result()

    try:
        # attempt to get the views shape manager
        shape_manager = view.GetCropRegionShapeManager()

        # check if the shape manager is None
        if shape_manager is None:
            return_value.update_sep(
                False,
                "Could not get the crop region shape manager for the view: {}.".format(
                    view
                ),
            )
            return return_value

        # set up an action to be executed in a transaction
        def action():
            action_return_value = Result()
            action_return_value =apply_view_crop_without_transaction_wrapper(view=view, crop_loop=crop_loop)
            return action_return_value

        transaction = Transaction(doc, "changing view crop")
        return_value = in_transaction(transaction, action)

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to update view crop: {} with error: {}".format(view, e),
        )

    return return_value


def apply_view_bounding_box_crop_without_transaction_wrapper(view, bbox):
    """
    Apply a crop region to a view.

    This method sets the crop region of the view to the specified crop loops.

    :param view: The view to apply the crop region to.
    :type view: Autodesk.Revit.DB.View
    :param bbox: The bounding box to apply as crop to the view.
    :type cbbox: Autodesk.Revit.DB.BoundingBoxXYZ

    :return: The result of the operation.
    :rtype: duHast.Utilities.Objects.result.Result
    """

    return_value = Result()

    try:
        # set the crop shape
        view.CropBox = bbox
        return_value .update_sep(
            True, "Successfully updated view crop: {}".format(view)
        )
           
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to update view crop: {} with error: {}".format(view, e),
        )

    return return_value


def apply_view_bounding_box_crop(doc, view, bbox):
    """
    Apply a crop region to a view.

    This method sets the crop region of the view to the specified crop loops.

    :param doc: The document containing the view.
    :type doc: Autodesk.Revit.DB.Document
    :param view: The view to apply the crop region to.
    :type view: Autodesk.Revit.DB.View
    :param bbox: The bounding box to apply as crop to the view.
    :type cbbox: Autodesk.Revit.DB.BoundingBoxXYZ

    :return: The result of the operation.
    :rtype: duHast.Utilities.Objects.result.Result
    """

    return_value = Result()

    try:

        # set up an action to be executed in a transaction
        def action():
            action_return_value = Result()
            action_return_value= apply_view_bounding_box_crop_without_transaction_wrapper(view, bbox)
            return action_return_value

        transaction = Transaction(doc, "changing view crop")
        return_value = in_transaction(transaction, action)

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to update view crop: {} with error: {}".format(view, e),
        )

    return return_value





"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit view schedules export to file. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_io import get_directory_path_from_file_path, get_file_name_without_ext, get_file_extension
from duHast.Utilities.directory_io import directory_exists


from Autodesk.Revit.DB import ExportColumnHeaders, ExportTextQualifier, ViewSchedule, ViewScheduleExportOptions

def export_schedule_to_file(schedule, file_path, field_delimiter=",", export_title=False, export_column_headers = ExportColumnHeaders.None ,export_text_qualifier=ExportTextQualifier.DoubleQuote):

    """
    Exports a Revit schedule to a file.

    :param schedule: The schedule to export.
    :type schedule: Autodesk.Revit.DB.ViewSchedule
    :param file_path: The file path to export the schedule to.
    :type file_path: str
    :param field_delimiter: The delimiter to use between fields in the exported file.
    :type field_delimiter: str, optional
    :param export_title: Whether to include the schedule title in the exported file.
    :type export_title: bool, optional
    :param export_column_headers: The type of column headers to include in the exported file.
    :type export_column_headers: Autodesk.Revit.DB.ExportColumnHeaders, optional
    :param export_text_qualifier: The text qualifier to use in the exported file.
    :type export_text_qualifier: Autodesk.Revit.DB.ExportTextQualifier, optional
    
    :return: Result object containing success status and messages.
    :rtype: duHast.Utilities.Objects.result.Result
    """

    return_value = Result()

    try:

        # check if the schedule is a view schedule instance
        if not isinstance(schedule, ViewSchedule):
            return_value.update_sep(False, "Provided schedule is not a valid ViewSchedule instance. Got {} instead.".format(type(schedule).__name__))
            return return_value

        # check if a file path was provided
        if not file_path:
            return_value.update_sep(False, "No file path provided for schedule export.")
            return return_value
        
        # check if file_path is a string
        if not isinstance(file_path, str):
            return_value.update_sep(False, "File path must be a string. Got {} instead.".format(type(file_path).__name__))
            return return_value


        # set up an schedule export options object with default values
        export_options = ViewScheduleExportOptions()
        export_options.FieldDelimiter = field_delimiter
        export_options.Title = export_title
        export_options.TextQualifier = export_text_qualifier
        export_options.ColumnHeaders = export_column_headers

        # get directory from file_path
        directory_path = get_directory_path_from_file_path(file_path)

        # check if the directory exists
        if not directory_exists(directory_path):
            return_value.update_sep(False, "Directory does not exist: {directory}".format(directory=directory_path))
            return return_value

        # get the file name from the file_path
        file_name = get_file_name_without_ext(file_path)
        # add the extension back in
        file_extension = get_file_extension(file_path)
        file_name += file_extension

        # export the schedule to a file
        schedule.Export(directory_path, file_name, export_options)

        return_value.append_message( "Schedule exported successfully to: {file}".format(file=file_path))
        return return_value

    except Exception as e:
        return_value.update_sep(False, "Failed to export schedule to file: {e}".format(e=e))
        return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Warnings solver utility class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import System
import clr

# from collections import namedtuple

from duHast.Utilities.Objects import result as res
from duHast.Revit.Warnings.warnings import get_warnings_by_guid
from duHast.Revit.Warnings.solver_room_tag_to_room import (
    RevitWarningsSolverRoomTagToRoom,
)
from duHast.Revit.Warnings.solver_duplicate_mark import RevitWarningsSolverDuplicateMark
from duHast.Revit.Warnings.solver_area_separation_lines_overlap import (
    RevitWarningsSolverAreaSepLinesOverlap,
)
from duHast.Revit.Warnings.solver_room_separation_lines_overlap import (
    RevitWarningsSolverRoomSepLinesOverlap,
)

# import Autodesk
# import Autodesk.Revit.DB as rdb

from duHast.Utilities.Objects import base

# a class used to return the value  if any, a message and the status of a method (true if everything is ok or false if something went wrong)
class RevitWarningsSolver(base.Base):

    # --------------------------- available filters ---------------------------

    def default_filter_return_all(self, doc, element_id):
        """
        Default filter for elements past into solver...returns True for any element past in.

        :param doc: Current Revit model document.
        :type doc: Autodesk.Revit.DB.Document
        :param element_id: The id of the element to be checked.
        :type element_id: Autodesk.Revit.DB.ElementId

        :return: True always.
        :rtype: bool
        """
        return True

    # --------------------------- solvers initialise code ---------------------------

    #: default solver classes
    solver_room_tag_to_room = RevitWarningsSolverRoomTagToRoom()
    solver_same_mark = RevitWarningsSolverDuplicateMark(default_filter_return_all)
    solver_room_sep_lines = RevitWarningsSolverRoomSepLinesOverlap()
    solver_area_sep_lines = RevitWarningsSolverAreaSepLinesOverlap()

    #: solver dictionary of available warning solvers by guid
    AVAILABLE_SOLVERS = {
        solver_room_tag_to_room.GUID: solver_room_tag_to_room,
        solver_same_mark.GUID: solver_same_mark,
        solver_room_sep_lines.GUID: solver_room_sep_lines,
        solver_area_sep_lines.GUID: solver_area_sep_lines,
    }

    # --------------------------- solvers code ---------------------------

    def __init__(self):
        """
        Constructor: assigns the default element filter (pass all)
        """

        # ini super class to allow multi inheritance in children!
        super(RevitWarningsSolver, self).__init__()

        self.filter_function_same_mark = self.default_filter_return_all

    def set_same_mark_filter_and_filter_solver(self, same_mark_filter_solver):
        """
        Method allowing to override the default filter function

        :param same_mark_filter_solver: A function to filter elements in warnings by
        :type same_mark_filter_solver: func(document, elementId, list of filter values)
        """

        # replace the old solver
        self.AVAILABLE_SOLVERS[same_mark_filter_solver.GUID] = same_mark_filter_solver
        self.solver_same_mark = same_mark_filter_solver

    def solve_warnings(self, doc):
        """
        Attempts to solve some warning in a revit model using available warnings solver.

        It will get all warnings in model, filter them by available solver GUIDs and finally will attempt to solve the warnings matched up with a solver.

        :param doc: Current Revit model document.
        :type doc: Autodesk.Revit.DB.Document
        :return:
            Result class instance.

            - .status True if all warnings could be solved. Otherwise False.
            - .message will contain all messages solver returned.

        :rtype: :class:`.Result`
        """

        return_value = res.Result()
        try:
            for solver in self.AVAILABLE_SOLVERS:
                warnings = get_warnings_by_guid(
                    doc, self.AVAILABLE_SOLVERS[solver].GUID
                )
                try:
                    result_solver = self.AVAILABLE_SOLVERS[solver].solve_warnings(
                        doc, warnings
                    )
                    return_value.update(result_solver)
                except Exception as e:
                    return_value.update_sep(
                        False,
                        "Failed to apply solver for warnings guid {} with exception: {}".format(
                            self.AVAILABLE_SOLVERS[solver].GUID, e
                        ),
                    )
        except Exception as e:
            return_value.update_sep(
                False, "Failed to apply all solvers with exception: {}".format(e)
            )
        return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Interface for family data storage class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


Notes:

- any revit specific data types i.e. ElementId need to be converted to int or string for JSON to work and in order for these classes to be used outside the Revit API!
- any string values need to be encoded to utf-8 and decoded back to utf-8 to avoid encoding issues


"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import base
from duHast.Utilities.utility import encode_utf8
from duHast.Revit.Family.Data.Objects.ifamily_data_storage_used_by import (
    IFamilyDataStorageUsedBy,
)
import System
import json


class IFamilyDataStorage(base.Base):
    def __init__(
        self,
        data_type,
        root_name_path,
        root_category_path,
        family_name,
        family_file_path,
        **kwargs
    ):

        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!
        super(IFamilyDataStorage, self).__init__(**kwargs)
        
        # set defaults
        self.data_type = None
        self.root_name_path = None
        self.root_category_path = None
        self.family_name = None
        self.family_file_path = None
        
        if isinstance(data_type, str):
            self.data_type = data_type
        else:
            raise ValueError("data_type must be a string")

        if isinstance(root_name_path, str):
            self.root_name_path = root_name_path
        else:
            raise ValueError("root_name_path must be a string")

        if isinstance(root_category_path, str):
            self.root_category_path = root_category_path
        else:
            raise ValueError("root_category_path must be a string")

        if isinstance(family_name, str):
            self.family_name = family_name
        else:
            raise ValueError("family_name must be a string")

        if isinstance(family_file_path, str):
            self.family_file_path = family_file_path
        else:
            raise ValueError("family_file_path must be a string")

    def __eq__(self, other):
        """
        Custom compare is equal override.

        :param other: Another instance of IFamilyDataStorage base class
        :type other: :class:`.IFamilyDataStorage`
        :return: True if all properties of compared class instances are equal, otherwise False.
        :rtype: Bool
        """

        return isinstance(other, IFamilyDataStorage) and (
            self.data_type,
            self.root_name_path,
            self.root_category_path,
            self.family_name,
            self.family_file_path,
        ) == (
            other.data_type,
            other.root_name_path,
            other.root_category_path,
            other.family_name,
            other.family_file_path,
        )

    # python 2.7 needs custom implementation of not equal
    def __ne__(self, other):
        return not self.__eq__(other=other)

    def update_usage(self, other_storage):
        """
        Update the usage of this storage object with the usage of another storage object by:

        - adding the use counter of the other storage object to this storage object's use counter
        - appending the root name path of the other storage object to this storage object's used_by list

        """

        if isinstance(other_storage, IFamilyDataStorage):
            self.use_counter = self.use_counter + other_storage.use_counter
            # just append the used by storage class
            for dummy in other_storage.used_by:
                self.used_by.append(dummy)
        else:
            raise ValueError(
                "other_storage is not of type IFamilyDataStorage but of type: {}".format(
                    type(other_storage)
                )
            )

    def _fix_data_types(self, value):
        """
        Replace any ElementId and Byte values with int or string respectively to have JSON working ok.
        Encodes any string in utf-8 and decodes it back to utf-8 to avoid encoding issues.
        Will do the same for list items!

        Any other type of values are not changed.

        :param flattened_dic: Dictionary of which values are to be converted.
        :type flattened_dic: {}
        :return: Dictionary with converted values.
        :rtype: {}
        """

        if isinstance(value, str):
            value = encode_utf8(value)
        elif isinstance(value, System.Byte):
            value = str(value)
        elif isinstance(value, list):
            value = [self._fix_data_types(item) for item in value]

        return value

    def get_data(self):
        """
        Get the data from the object als dictionary.

        Note data type fixing is applied to:

        - ElementIds
        - System.Byte values
        """

        dict = {}
        for key, value in self.__dict__.items():
            # only add non private properties
            if not key.startswith("_"):
                dict[key] = self._fix_data_types(value)
        return dict

    def get_data_values_as_list_of_strings(self):
        data_list = []
        for key, value in self.__dict__.items():
            # set a default value
            value_updated = "None"
            if isinstance(value, list):
                # check if a value is a list of IFamilyDataStorageUsedBy objects
                # if so convert them to list of json dictionaries
                if len(value) > 0:
                    # check if list of used by instances
                    if isinstance(value[0], IFamilyDataStorageUsedBy):
                        # list of used by instances
                        list_string = [item.to_json() for item in value]
                        value_updated = "[{}]".format(",".join(list_string))
                    else:
                        # list of standard values
                        # encode all values to utf-8
                        # System.Byte to string, ElementId to int
                        value_updated = self._fix_data_types(value)
                        # convert individual values to string
                        list_string = [str(item) for item in value_updated]
                        value_updated = "[{}]".format(",".join(list_string))
            else:
                # encode all values to utf-8
                # System.Byte to string, ElementId to int
                value_updated = self._fix_data_types(value)
                if (
                    isinstance(value_updated, str) == False
                    and isinstance(value_updated, list) == False
                ):
                    value_updated = str(value_updated)
            # add to data list
            data_list.append(value_updated)
        return data_list

    def get_property_names(self):
        names = []
        for key in self.__dict__.keys():
            if not key.startswith("_"):
                names.append(key)
        return names

    def to_json(self):
        """
        Convert the instance of this class to json.

        :return: A Json object.
        :rtype: json
        """

        data_cleaned = self.get_data()
        json_object = json.dumps(
            dict(data_cleaned), indent=None, default=lambda o: o.__dict__
        )
        return json_object



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit view port to data view port conversion. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Data.Objects.Collectors.data_sheet_view_port import DataSheetViewPort
from duHast.Data.Objects.Collectors.Properties.data_view_port_type_names import (
    DataViewPortTypeNames,
)
from duHast.Data.Objects.Collectors.data_view_3d import DataViewThreeD
from duHast.Data.Objects.Collectors.data_view_elevation import DataViewElevation
from duHast.Data.Objects.Collectors.data_view_plan import DataViewPlan
from duHast.Data.Objects.Collectors.data_view_schedule import DataViewSchedule
from duHast.Data.Objects.Collectors.Properties.data_schedule_segement import DataScheduleSegment


from duHast.Utilities.unit_conversion import convert_imperial_feet_to_metric_mm

from duHast.Revit.Common.Geometry.points import convert_XYZ_to_point2,convert_XYZ_to_point3
from duHast.Revit.Views.elevation import get_view_index_on_marker

from Autodesk.Revit.DB import SectionType, ViewType


def _get_view_port_type(doc, revit_view_port):
    """
    Returns a string reprensenting the viewport type on None if to be ignored (not match with view port types of interest)

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param view_port: A Revit ViewPort
    :type view_port: Autodesk.Revit.DB.ViewPort

    :return: The viewport type name or none
    :rtype: str or None
    """

    view = doc.GetElement(revit_view_port.ViewId)

    if view.ViewType == ViewType.FloorPlan:
        return DataViewPortTypeNames.FLOOR_PLAN
    elif view.ViewType == ViewType.Elevation:
        return DataViewPortTypeNames.ELEVATION
    elif view.ViewType == ViewType.ThreeD:
        return DataViewPortTypeNames.THREE_D
    elif view.ViewType == ViewType.Schedule:
        # thats unlikely
        return DataViewPortTypeNames.SCHEDULE
    else:
        return None


def _get_view_crop(view):
    # get bounding box
    crop_box = view.CropBox

    # there is a very very slight chance that view is not cropped ...it should be but...
    if crop_box:
        # get the outlines min and max points as 2d points
        bb_max_2d = convert_XYZ_to_point2(crop_box.Max)
        bb_min_2d = convert_XYZ_to_point2(crop_box.Min)

        return bb_min_2d, bb_max_2d
    else:
        return None


def _get_plan_view(view):
    """
    Converts data from a Revit plan view to a data plan view instance

    :param view: The Revit view
    :type view: Autodesk.Revit.DB.ViewPlan

    :return: A data view instance
    :rtype: :class:`.DataViewPlan`
    """

    data_instance = DataViewPlan()

    # get the view crop
    bb_min_2d, bb_max_2d = _get_view_crop(view)
    if bb_min_2d and bb_max_2d:
        data_instance.bounding_box.update(bb_min_2d, bb_max_2d)

    # get any tags in the view
    return data_instance

def _get_elevation_view(doc, view):
    """
    Converts data from a Revit elevation view to a data elevation view instance

    :param view: The Revit view
    :type view: Autodesk.Revit.DB.ViewSection

    :return: A data view instance
    :rtype: :class:`.DataViewElevation`
    """

    data_instance = DataViewElevation()

    # get bounding box
    bb_min_2d, bb_max_2d = _get_view_crop(view)
    if bb_min_2d and bb_max_2d:
        data_instance.bounding_box.update(bb_min_2d, bb_max_2d)

    marker = get_view_index_on_marker(doc, view)
    # store the index
    data_instance.marker_index = marker

    # orientation (eye point and view direction)
    data_instance.view_direction = convert_XYZ_to_point3(view.ViewDirection)

    return data_instance


def _get_three_d_view(view):
    """
    Converts data from a Revit 3D view to a data 3D view instance

    :param view: The Revit view
    :type view: Autodesk.Revit.DB.View3D

    :return: A data view instance
    :rtype: :class:`.DataViewThreeD`
    """

    data_instance = DataViewThreeD()

    # get bounding box
    bb_min_2d, bb_max_2d = _get_view_crop(view)
    if bb_min_2d and bb_max_2d:
        data_instance.bounding_box.update(bb_min_2d, bb_max_2d)

    # orientation (eye point and view direction)
    view_orientation = view.GetOrientation()
    data_instance.eye_position = convert_XYZ_to_point3(view_orientation.EyePosition)
    data_instance.forward_direction = convert_XYZ_to_point3(view_orientation.ForwardDirection)
    data_instance.up_direction = convert_XYZ_to_point3(view_orientation.UpDirection)

    return data_instance


def _get_schedule_view(view):
    """
    Converts data from a Revit schedule view to a data schedule view instance

    :param view: The Revit view
    :type view: Autodesk.Revit.DB.ViewSchedule

    :return: A data view instance
    :rtype: :class:`.DataViewSchedule`
    """

    data_instance = DataViewSchedule()

    # here is a way to extract number of rows: (old...)
    # https://thebuildingcoder.typepad.com/blog/2012/05/the-schedule-api-and-access-to-schedule-data.html
    # this might be better and more up to date: https://forums.autodesk.com/t5/revit-api-forum/how-to-get-schedule-data/td-p/7319520
    # in fact it does show how to get to number of rows easily...

    table = view.GetTableData()
    section = table.GetSectionData(SectionType.Body)
    number_of_rows = section.NumberOfRows

    # store the number of data rows
    data_instance.total_number_of_rows = number_of_rows

    # The total count of schedule segments. 1 means the schedule is not split yet.
    for i in range(view.GetSegmentCount() - 1):
        seg_data = DataScheduleSegment()
        seg_data.index = i
        seg_height = convert_imperial_feet_to_metric_mm(view.GetSegmentHeight(i))
        seg_data.height = seg_height
        data_instance.segments.append(seg_data)

    # get bounding box
    # TODO:
    return data_instance


def _get_view_data(doc, view):
    """
    Set up view data instance depending ov view type

    :param view: The view to be converted
    :type view: Autodesk.Revit.DB.View

    :return: A data view instance
    :rtype: :class:`.DataViewBase`
    """

    view_data_instance = None

    # check view type
    if view.ViewType == ViewType.FloorPlan:
        # plan view
        view_data_instance = _get_plan_view(view=view)
    elif view.ViewType == ViewType.Elevation:
        # elevation
        view_data_instance = _get_elevation_view(doc=doc, view=view)
    elif view.ViewType == ViewType.ThreeD:
        # 3D
        view_data_instance = _get_three_d_view(view=view)
    elif view.ViewType == ViewType.Schedule:
        # schedule
        view_data_instance = _get_schedule_view(view=view)

    return view_data_instance


def convert_revit_viewport_to_data_instance(doc, revit_view_port):
    """
    Converts a Revit ViewPort into a data viewport

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param view_port: A Revit ViewPort
    :type view_port: Autodesk.Revit.DB.ViewPort
    :return: A populated data viewport instance
    :rtype: :class:`.DataSheetViewPort`
    """

    view_port_type = _get_view_port_type(doc, revit_view_port)
    if view_port_type == None:
        # ignore this viewport
        return None

    # set up data instances
    view_port_data = DataSheetViewPort()

    # get an outline from the Revit view port
    view_port_outline = revit_view_port.GetBoxOutline()

    # get the outlines min and max points as 2d points
    bb_max_2d = convert_XYZ_to_point2(view_port_outline.MaximumPoint)
    bb_min_2d = convert_XYZ_to_point2(view_port_outline.MinimumPoint)

    # update the bounding box property of the view port instance
    view_port_data.bounding_box.update(bb_min_2d, bb_max_2d)

    # get the viewport centre
    centre_point = convert_XYZ_to_point2(revit_view_port.GetBoxCenter())
    view_port_data.centre_point = centre_point

    # set the viewport type
    view_port_data.vp_type = view_port_type

    # set the view
    revit_view = doc.GetElement(revit_view_port.ViewId)
    view_data = _get_view_data(doc=doc, view=revit_view)
    view_port_data.view = view_data

    return view_port_data


def convert_revit_schedule_sheet_instances_to_data_instance(
    doc, sheet, revit_schedule_sheet_instances
):
    """
    Converts a list of schedule sheet instances to view port data instances

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param sheet: The sheet of which to get the schedule instance from.
    :type sheet: Autodesk.Revit.DB.ViewSheet
    :param revit_schedule_sheet_instances: A list of schedule sheet instances
    :type revit_schedule_sheet_instances: [Autodesk.Revit.DB.ScheduleSheetInstance]
    :return: A list of populated data viewport instances
    :rtype: [:class:`.DataSheetViewPort`]
    """

    all_view_port_data = []

    # loop over all schedule sheet instances
    for schedule_sheet_instance in revit_schedule_sheet_instances:

        # set up data instances
        view_port_data = DataSheetViewPort()
        view_port_data.vp_type = DataViewPortTypeNames.SCHEDULE

        # bounding box:
        bbox = schedule_sheet_instance.get_BoundingBox(sheet)
        # get the outlines min and max points as 2d points
        bb_max_2d = convert_XYZ_to_point2(bbox.Max)
        bb_min_2d = convert_XYZ_to_point2(bbox.Min)

        # get the viewport centre
        centre_point = convert_XYZ_to_point2(schedule_sheet_instance.Point)
        view_port_data.centre_point = centre_point

        # update the bounding box property of the view port instance
        view_port_data.bounding_box.update(bb_min_2d, bb_max_2d)

        # get the schedule view data
        schedule_view = doc.GetElement(schedule_sheet_instance.ScheduleId)
        view_data = _get_schedule_view(schedule_view)
        view_port_data.view = view_data

        # append to list to be returned
        all_view_port_data.append(view_port_data)

    return all_view_port_data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around data storage in Revit.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Data storage can be used to project wide settings not associated with a specific element. 

Obsolete in Revit 2025 ... 
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

import duHast.Utilities.Objects.result as res
from duHast.Revit.Common.transaction import in_transaction


from Autodesk.Revit.DB import Transaction
from Autodesk.Revit.DB.ExtensibleStorage import DataStorage, Entity


def create_project_data_storage(doc, schema):
    """
    Create a new DataStorage element in the Revit document and set its schema.
    
    :param doc: The Revit document to create the DataStorage element in.
    :type doc: Autodesk.Revit.DB.Document
    
    :param schema: The schema to set on the DataStorage element.
    :type schema: Autodesk.Revit.DB.ExtensibleStorage.Schema
    
    :return: A Result object containing the created DataStorage element in its result list or an error message.
    :rtype: duHast.Utilities.Objects.result.Result
    """
    
    return_value = res.Result()
    
    # build an action to create a new DataStorage element
    def action():
        action_return_value = res.Result()
        try:
            # Create a new DataStorage element
            data_storage = DataStorage.Create(doc)
            entity = Entity(schema)
            # Set the schema to the DataStorage element
            data_storage.SetEntity(entity)
            action_return_value.append_message(
                "DataStorage element created successfully."
            )
            action_return_value.result.append(data_storage)
        except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to set up data storage: {}".format(e),
                )
        return action_return_value
    
    # attempt to create the DataStorage element in a transaction
    transaction = Transaction(doc, "Creating data storage")
    data_storage_creation_result = in_transaction(transaction, action)
    return_value.update(data_storage_creation_result)

    return return_value




"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit materials helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

# import common library modules

# import Autodesk
import Autodesk.Revit.DB as rdb

# --------------------------------------------- utility functions ------------------


def get_all_materials(doc):
    """
    Gets all materials in a model.

    Filter by class.

    :param doc: _description_
    :type doc: _type_

    :return: A filtered element collector of materials
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rdb.FilteredElementCollector(doc).OfClass(rdb.Material)
    return collector


def get_material_by_id(doc, id):
    """
    Gets a material element based on a material id.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param id: The id of material to be returned.
    :type id: Autodesk.Revit.DB.ElementId

    :return: A material if matching id was found. Otherwise nothing gets returned!
    :rtype: Autodesk.Revit.DB.Material
    """

    mats = get_all_materials(doc)
    for m in mats:
        if m.Id.IntegerValue == id.IntegerValue:
            return m


def get_material_name_by_id(doc, id):
    """
    Gets a material name based on a material id.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param id: Id of material of which the name is to be returned.
    :type id: Autodesk.Revit.DB.ElementId

    :return: The material name if matching id was found or the default value: '<By Category>'
    :rtype: str
    """

    name = "<By Category>"
    mats = get_all_materials(doc)
    for m in mats:
        if m.Id.IntegerValue == id.IntegerValue:
            m_name = rdb.Element.Name.GetValue(m)
            name = "" if m_name == None else m_name
    return name



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
An implementation of a custom wpf window base class which can be shown in the Revit context.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
clr.AddReference("PresentationFramework")

from System.Windows import Window, VerticalAlignment, HorizontalAlignment
from duHast.UI.Objects.WPF.Xaml.XamlLoader import XamlLoader
from duHast.UI.Objects.WPF.Exceptions.MissingXAMLException import MissingXAMLException

class WPFWindowBase(Window):
    def __init__(self, xaml_path, main_view_model, xaml_by_view_model, resources_xaml_path):
    
        # make sure to initialize the base class
        super(WPFWindowBase, self).__init__()
        
        # load start up xaml data (main window)
        xaml_loader = XamlLoader(xaml_path)
        
        # get the xaml root
        xaml_window = xaml_loader.Root
        
        # get the main grid control from main window
        # this control will host all other views and associated view models
        self.main_grid = xaml_loader.MainGrid
        
        # Ensure xaml_window is of type Window and use it
        if isinstance(xaml_window, Window):
            self.Content = xaml_window.Content
            self.Title = xaml_window.Title
            self.Width = xaml_window.Width
            self.SizeToContent = xaml_window.SizeToContent
        else:
            self.Content = xaml_window

        # set up any static styling resources
        # in c# land these would be added in App.xaml
        if( resources_xaml_path):
            resources_loader = XamlLoader(resources_xaml_path)
            self.Resources.MergedDictionaries.Add(resources_loader.Root)
        
        # store xaml views by view model
        self.xaml_path_by_view_model = xaml_by_view_model
        
        # Store the passed-in MainViewModel
        self.main_view_model = main_view_model
        
        # Set the DataContext of the window to MainViewModel
        self.DataContext = self.main_view_model
        
        # Bind the CurrentViewModel to the ContentControl
        self.update_content_view(self.main_view_model.CurrentViewModel)
        
        # Subscribe to changes in CurrentViewModel
        self.main_view_model._navigation_store.add_ViewModelChanged (self.on_current_view_model_changed)
        
        # Wire up the Closed event
        self.Closed += self.on_closed
    
    
    def on_current_view_model_changed(self):
        # Update the view when CurrentViewModel changes
        self.update_content_view(self.main_view_model.CurrentViewModel)


    def update_content_view(self, view_model):
        """
        Dynamically set the content of the main grid control based on the current view model

        Args:
            view_model (_type_): The new view model to be applied
        """
        
        # get the xaml path depending on the view model
        xaml_path_current_view = self.xaml_path_by_view_model.get(type(view_model),None)

        # Load the XAML file if it exists
        if xaml_path_current_view :
            # load xaml data matching view model
            xaml_loader_current_view = XamlLoader(xaml_path_current_view)
            
            # set the data context to the new view model
            xaml_loader_current_view.Root.DataContext = view_model
            
            # Clear the existing children in the main grid
            self.main_grid.Children.Clear()

            # Set alignment for full stretching (not sure whether this is doing anything??)
            xaml_loader_current_view.VerticalAlignment = VerticalAlignment.Stretch
            xaml_loader_current_view.HorizontalAlignment = HorizontalAlignment.Stretch

            # Add the new view to the main grid
            self.main_grid.Children.Add(xaml_loader_current_view.Root)
        else:
            raise MissingXAMLException(message="XAML file not provided for view model.", view_model=view_model)
            
        
    def on_closed(self, sender, event):
        """
        Handles the Closed event.
        
        :param sender: The sender.
        :type sender: object
        :param event: The event.
        :type event: object
        """
        # Call the action on close method
        self.action_on_close()
    
    def action_on_close(self):
        """
        Handles the action on close.
        
        Can be overwritten by derived classes to handle cleanup tasks.
        """
        # Handle cleanup here
        pass


# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Common import parameter_get_utils as rParaGet

def get_sheet_parameter_data (view_sheet):
    """
    Retrieves the parameter data from a given view sheet.

    :param view_sheet: The view sheet element from which to retrieve parameter data.
    :type view_sheet: Autodesk.Revit.DB.ViewSheet
    :return: A dictionary containing the parameter names and their corresponding values.
    :rtype: dict
    """

    data = {}
    paras = view_sheet.GetOrderedParameters()

    for para in paras:
            # get values as utf-8 encoded strings
            # for some characters this still throws an exception...added ascii encoding
            value = rParaGet.get_parameter_value_utf8_string(para)
            try:
                data[para.Definition.Name] = value
            except:
                data[para.Definition.Name] = "Failed to retrieve value"

    return data

def get_naming_chunks(sheet_name_string):
    """
    Splits the sheet name string into chunks based on the delimiters.
    
    :param sheet_name_string: The sheet name string to split.
    :type sheet_name_string: str
    :return: A list of chunks.
    :rtype: list
    """
    
    # Split the string by the delimiters and return the chunks
    return [chunk.strip() for chunk in sheet_name_string.split("*") if chunk.strip()]


def replace_illegal_characters_from_dwg_file_name(current_file_name, replace_full_stop=True):
    """
    replaces illegal characters with dashes in the given string.
    
    :param string: The string to replace characters in.
    :type string: str
    
    :return: The amended string.
    :rtype: str
    """
    
    # Define the illegal characters
   
    illegal_characters = {
        ".":"-",
        "/":"-",
    }
    
    # Replace each illegal character with the designated replacement string
    for char, replacement in illegal_characters.items():
        # only replace full stop if indicated
        if char == "." and not replace_full_stop:
            continue
        current_file_name = current_file_name.replace(char,  replacement)
    
    return current_file_name


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit rooms tags. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import Element, Transaction
from duHast.Revit.Common.transaction import in_transaction
from duHast.Utilities.Objects import result as res


def move_tag_to_room(doc, tag_id, transaction_manager=in_transaction):
    """
    Moves a room tag to the associated rooms location point. This will

    - un-pin the tag
    - disable any leader

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param tag_id: The element id of the tag to be moved to the room.
    :type tag_id: Autodesk.Revit.DB.ElementId
    :param transaction_manager: The transaction manager used to perform the modifications.
    :type transaction_manager: (function)
    :return: Result class instance.

        - Tag moving status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the name and number of the room.

        On exception (handled by optimizer itself!):

        - result.status (bool) will be False.
        - result.message will contain the name and number of the room and the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    rt = doc.GetElement(tag_id)
    room_tag_point = rt.Location.Point
    room_location_point = rt.Room.Location.Point
    room_data = "{} {}".format(rt.Room.Number, Element.Name.GetValue(rt.Room))
    translation = room_location_point - room_tag_point

    def action():
        action_return_value = res.Result()
        try:
            # check if tag is pinned...
            if rt.Pinned:
                rt.Pinned = False

            # disable leader if there is one
            if rt.HasLeader:
                rt.HasLeader = False

            # and move the tag
            rt.Location.Move(translation)
            action_return_value.append_message ("Moved tag to room: {}".format(room_data))
        except Exception as e:
            action_return_value.update_sep(
                False,
                "Failed to move tag to room: {} with exception: {}".format(
                    room_data, e
                ),
            )
        return action_return_value

    transaction = Transaction(doc, "Moving room tag to room : {}".format(room_data))
    return_value.update(transaction_manager(transaction, action))
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Sample showing how to find which ceilings are in which rooms.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module requires python >3.9 due to dependencies:

    - numpy
    - shapely

This module:

    - collects ceiling and room data instances by level ( assume a ceiling is always modelled as the room it is in )
    - converts room and ceiling outlines to shapely polygons
    - test for intersection of all ceilings on a given level with all rooms on a given level
    - stores any intersections found ( does a check how much area is  intersecting...if to small its assumed its not an intended intersection)
    - reports all rooms and any associated ceiling(s) found

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import codecs
import csv


from duHast.Utilities.Objects import result as res
from duHast.Data.Objects.Collectors import data_ceiling as dc
from duHast.Data.Objects.Collectors import data_room as dr
from duHast.Data.Utils import data_import as dReader
from duHast.Data import data_to_shapely as dToS

# TODO: check if this module works in python 3.x
# if not, raise an exception
try:
    from duHast.Utilities.files_csv import write_report_data_as_csv
except ImportError as e:
    raise ImportError(
        "This module requires the duHast.Utilities.files_csv module to be available. Error: {}".format(
            e
        )
    )

# --------------- writing out data ------------------


def _read_data(file_path):
    """
    Reads text files into data objects within data reader class which is returned

    Data file to be json formatted. (one json entry per row)

    :param filePath: Fully qualified path to json formatted data file.
    :type filePath: str

    :return: A file data reader instance.
    :rtype: :class:`.ReadDataFromFile`
    """

    # read json file and convert into data objects
    dataReader = dReader.ReadDataFromFile(file_path)
    dataReader.load_data()
    return dataReader


# --------------- writing out data ------------------


def _write_report_data(
    file_name,
    header,
    data,
):
    """
    Method writing out report information to csv file.

    :param file_name: Fully qualified file path to data file.
    :type file_name: str
    :param header: List of column headers, provide empty list if not required!
    :type header: list[str]
    :param data: List of list of strings representing row data
    :type data: list[list[str]]

    :return:
        Result class instance.

        - result.status False if an exception occurred, otherwise True.
        - result.message will contain file name of file written.
        - result.result: empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.
        - result.result: will be an empty list

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        # open the file in the write mode
        with codecs.open(file_name, "w", encoding="utf-8") as f:
            # create the csv writer
            writer = csv.writer(f)
            # check header
            if len(header) > 0:
                writer.writerow(header)
            if len(data) > 0:
                for d in data:
                    # write a row to the csv file
                    writer.writerow(d)
            f.close()
        return_value.update_sep(
            True, "Successfully wrote data to: {}".format(file_name)
        )
    except Exception as e:
        return_value.update_sep(False, "Failed to write data to: {}".format(file_name))
    return return_value


# --------------- data processing ------------------


def _build_dictionary_by_level_and_data_type(data_reader):
    """
    Returns a dictionary where:

    - key: is the level name
    - values is a list of list of data objects
    -       first list room data objects
    -       second list ceilings data objects

    :param data_reader: A data reader class instance
    :type data_reader: :class:`.ReadDataFromFile`

    :return: A dictionary where key is the level name, value is a tuple of two lists: first one are rooms, second ones are ceiling data objects.
    :rtype: dic{str:[[:class:`.DataRoom],[:class:`.DataCeiling`]]}
    """

    dic = {}
    for d_object in data_reader.data:
        if d_object.level.name not in dic:
            rooms_by_level = data_reader.get_data_by_level_and_data_type(
                d_object.level.name, dr.DataRoom.data_type
            )
            ceilings_by_level = data_reader.get_data_by_level_and_data_type(
                d_object.level.name, dc.DataCeiling.data_type
            )
            dic[d_object.level.name] = (rooms_by_level, ceilings_by_level)
    return dic


def _get_property_values_as_list(properties, property_keys):
    """
    _summary_

    :param properties: _description_
    :type properties: _type_
    :param property_keys: _description_
    :type property_keys: _type_

    :return: _description_
    :rtype: _type_
    """

    values = []
    for property_key in property_keys:
        if property_key in properties:
            values.append(str(properties[property_key]))
        else:
            values.append("null")
    return values


def _convert_object_data_into_report_data(
    dic_object,
    room_instance_property_keys,
    ceiling_type_property_keys,
    ceiling_instance_property_keys,
):
    """
    Converts a dictionary of DataRoom objects by level into list of lists of data entries per room so it can be written to file.

    :param dic_object:  A dictionary where key is the level name and values is a list of DataRoom instances.
    :type dic_object: {str:[:class: `.DataRoom`]}

    :return: List of list of strings representing room data
    :rtype: list of list [str]
    """

    data = []
    for levelName in dic_object:
        for room in dic_object[levelName][0]:
            # put fixed (always exported ) values first
            data_row = [
                room.level.name,
                room.revit_model.name,
                str(room.instance_properties.id),
                room.design_set_and_option.set_name,
                room.design_set_and_option.option_name,
                str(room.design_set_and_option.is_primary),
            ]
            # get custom values
            data_row = data_row + _get_property_values_as_list(
                room.instance_properties.properties, room_instance_property_keys
            )

            # data.append (data_row)

            associated_data_rows = []
            for associated_element in room.associated_elements:
                # only add ceiling data for now
                if associated_element.data_type == dc.DataCeiling.data_type:
                    associated_data_row = [
                        associated_element.level.name,
                        room.revit_model.name,
                        str(associated_element.instance_properties.id),
                        associated_element.design_set_and_option.set_name,
                        associated_element.design_set_and_option.option_name,
                        str(associated_element.design_set_and_option.is_primary),
                    ]

                    associated_data_row = (
                        associated_data_row
                        + _get_property_values_as_list(
                            associated_element.type_properties.properties,
                            ceiling_type_property_keys,
                        )
                    )
                    associated_data_row = (
                        associated_data_row
                        + _get_property_values_as_list(
                            associated_element.instance_properties.properties,
                            ceiling_instance_property_keys,
                        )
                    )
                    associated_data_rows.append(associated_data_row)

                    data.append(data_row + associated_data_row)
    return data


def _intersect_ceiling_vs_room(
    ceiling_poly_id,
    ceiling_polygon,
    room_poly_id,
    room_polygon,
    data_objects,
    level_name,
):
    """
    Does an intersection check of a ceiling polygon with a room polygon. If there is an intersection, the ceiling object will be added to the associated elements list of the room object.

    Note:

    - To avoid false positives: only ceiling which overlap a room by an area greater then 0.1 percent of the ceiling area will be considered as intersecting.
    - Manipulates the past in data_objects by reference!

    :param ceiling_poly_id: The Revit ceiling element id.
    :type ceiling_poly_id: int
    :param ceiling_polygon: A polygon representing the ceiling element.
    :type ceiling_polygon: shapely.polygon
    :param  room_poly_id: The Revit room element id.
    :type  room_poly_id: int
    :param room_polygon: A polygon representing the room element.
    :type room_polygon: shapely.polygon
    :param data_objects: A dictionary where key is the level name, value is a tuple of two lists: first one are rooms, second ones are ceiling data objects.
    :type data_objects: [str:([:class:`.DataRoom`],[:class:`.DataCeiling`])]
    :type  level_name: The level name where room and ceiling are on.
    :param room_polygon: str

    :return:
        Result class instance.

        - result.status False if an exception occurred, otherwise True.
        - result.message will contain processing messages.
        - result.result: empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.
        - result.result will be an empty list

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # add some exception handling here in case intersect check throws an error
    try:
        # check if a polygon is empty...if so ignore
        if ceiling_polygon.is_empty or room_polygon.is_empty:
            raise Exception(
                "Ceiling {} or room {} polygon is empty. Ignored!".format(
                    ceiling_poly_id, room_poly_id
                )
            )

        # check what exactly is happening
        if ceiling_polygon.intersects(room_polygon):
            # calculates percentage of overlapping ceiling area vs room area
            # anything less then 0.1 will be ignored...
            area_intersection_percentage_of_ceiling_vs_room = (
                ceiling_polygon.intersection(room_polygon).area / room_polygon.area
            ) * 100
            # check what percentage the overlap area is...if less then 0.1 percent ignore!
            if area_intersection_percentage_of_ceiling_vs_room < 0.1:
                # ceiling overlap area is to small...not in room
                return_value.append_message(
                    "Ceiling {} has an overlap of {} % with  room {}. Ignored!".format(
                        ceiling_poly_id,
                        area_intersection_percentage_of_ceiling_vs_room,
                        room_poly_id,
                    )
                )
            else:
                # ceiling is within the room: add to room data object
                # get the room object by its Revit ID
                data_object_room = list(
                    filter(
                        lambda x: (x.instance_properties.id == room_poly_id),
                        data_objects[level_name][0],
                    )
                )[0]
                # get the ceiling object by its Revit id
                data_object_ceiling = list(
                    filter(
                        lambda x: (x.instance_properties.id == ceiling_poly_id),
                        data_objects[level_name][1],
                    )
                )[0]
                # add ceiling object to associated elements list of room object
                data_object_room.associated_elements.append(data_object_ceiling)
                return_value.append_message(
                    "Added ceiling {} to room {}".format(ceiling_poly_id, room_poly_id)
                )
    except Exception as e:
        # get the offending elements:
        data_object_room = list(
            filter(
                lambda x: (x.instance_properties.id == room_poly_id),
                data_objects[level_name][0],
            )
        )[0]
        # set some predefined values in case the room object is not found
        room_name = "unknown"
        room_number = "unknown"
        room_id = "unknown"
        # populate values if room object is found
        if data_object_room:
            room_name = data_object_room.instance_properties.properties["Name"]
            room_number = data_object_room.instance_properties.properties["Number"]
            room_id = data_object_room.instance_properties.id

        data_object_ceiling = list(
            filter(
                lambda x: (x.instance_properties.id == ceiling_poly_id),
                data_objects[level_name][1],
            )
        )[0]
        # set some predefined values in case the ceiling object is not found
        ceiling_id = "unknown"
        # populate values if ceiling object is found
        if data_object_ceiling:
            ceiling_id = data_object_ceiling.instance_properties.id

        return_value.append_message(
            "Exception: {} \noffending room: room name: {} , room number: {} , room id: {} , \n...is valid polygon: {}\n...is empty polygon: {}\noffending ceiling id: {} , \n...is valid polygon: {}\n...is empty polygon: {}".format(
                e,
                room_name,
                room_number,
                room_id,
                room_polygon.is_valid,
                room_polygon.is_empty,
                ceiling_id,
                ceiling_polygon.is_valid,
                ceiling_polygon.is_empty,
            )
        )

    return return_value


# --------------- main functions ------------------


def write_data_to_file(
    data,
    output_file_path,
    room_instance_property_keys=["Number", "Name"],
    ceiling_type_property_keys=["Type Mark"],
    ceiling_instance_property_keys=["Height Offset From Level"],
):
    """
    Writes Room data to file.

    Note:

    - Room data consists of fixed (always reported) instance properties and custom instance properties defined in room_instance_property_keys
    - Ceiling data consists of fixed (always reported) instance properties, custom type properties defined in ceiling_type_property_keys and custom instance properties defined in ceiling_instance_property_keys
    - the report contains a row per associated element per room ( if 2 ceilings are in a room, the report will contain 2 rows)

    :param data: A dictionary where key is the level name and values is a list of DataRoom instances.
    :type data: {str:[:class: `.DataRoom`]}
    :param output_file_path: Fully qualified file path to output report file.
    :type output_file_path: str
    :param room_instance_property_keys: Names of room instance properties to be reported, defaults to ['Number', 'Name']
    :type room_instance_property_keys: list, optional
    :param ceiling_type_property_keys: Names of ceiling type properties to be reported, defaults to ['Type Mark']
    :type ceiling_type_property_keys: list, optional
    :param ceiling_instance_property_keys: Names of ceiling instance properties to be reported, defaults to ['Height Offset From Level']
    :type ceiling_instance_property_keys: list, optional

    :return:
        Result class instance.

        - Export status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the processing messages.
        - result.result will be an empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        converted_data = _convert_object_data_into_report_data(
            data,
            room_instance_property_keys,
            ceiling_type_property_keys,
            ceiling_instance_property_keys,
        )

        # create header for report file
        # rooms always written values first
        data_header = [
            "room level name",
            "room model name",
            "room revit id",
            "room design set name",
            "room design option name",
            "room design option is primary",
        ]
        # rooms custom data next
        data_header = data_header + room_instance_property_keys
        # ceilings always written values first
        data_header = data_header + [
            "ceiling level name",
            "ceiling model name",
            "ceiling revit id",
            "ceiling design set name",
            "ceiling design option name",
            "ceiling design option is primary",
        ]
        # ceilings custom data next
        data_header = (
            data_header + ceiling_type_property_keys + ceiling_instance_property_keys
        )
        # write data to file
        return_value.update(
            write_report_data_as_csv(output_file_path, data_header, converted_data)
            #_write_report_data(output_file_path, data_header, converted_data)
        )

    except Exception as e:
        return_value.update_sep(
            False, "Failed to write report with exception {}".format(e)
        )
    return return_value


def get_ceilings_by_room(data_source_path):
    """
    Reads Revit data from file and runs an intersection check of each ceiling on a level with each room on the same level.

    Note:
    DataRoom instance will contain any ceilings in that room in associated elements property.

    :param data_source_path: Fully qualified file path to json formatted file containing DataRoom and DataCeiling objects.
    :type data_source_path: str

    :return:
        Result class instance.

        - result.status False if an exception occurred, otherwise True.
        - result.message will contain processing messages.
        - result.result A dictionary where key is the level name, value is a tuple of two lists: first one are rooms, second ones are ceiling data objects.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain the exception message.
        - result.result will be an empty list

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # read exported ceiling and room data from file
    data_reader = _read_data(data_source_path)
    # check if read returned anything
    if len(data_reader.data) == 0:
        return_value.update_sep(
            False, "File: {} did not contain any valid data.".format(data_source_path)
        )
        return return_value

    print("\n".join(data_reader.debug_messages))

    # build dictionary of objects by level and object type
    data_objects = _build_dictionary_by_level_and_data_type(data_reader)
    return_value.append_message(
        "Data loaded contains {} levels: \n{}".format(
            len(data_objects),
            "...\n".join(
                [
                    "{}: {} entries, Rooms: {}, Ceilings: {}".format(
                        key, len(value[0]) + len(value[1]), len(value[0]), len(value[1])
                    )
                    for key, value in data_objects.items()
                ]
            ),
        )
    )
    # key level name, value tuple ( rooms [index 0] and ceilings [index 1])
    # loop over dic and process each key (level):
    #       - check if rooms and ceilings
    #       - intersection check
    #       - update room object with ceiling match
    # print(data_objects)
    for level_name in data_objects:
        # check rooms are on this level
        if len(data_objects[level_name][0]) > 0:
            # check ceilings are on this level
            if len(data_objects[level_name][1]) > 0:
                polygons_by_type = {}
                # convert geometry data off all rooms and ceilings into dictionaries : key is Revit element id, values are shapely polygons
                room_polygons = dToS.get_shapely_polygons_from_geo_object(
                    data_objects[level_name][0], dr.DataRoom.data_type
                )
                ceiling_polygons = dToS.get_shapely_polygons_from_geo_object(
                    data_objects[level_name][1], dc.DataCeiling.data_type
                )
                polygons_by_type[dr.DataRoom.data_type] = room_polygons
                polygons_by_type[dc.DataCeiling.data_type] = ceiling_polygons
                # loop over rooms ids and find intersecting ceilings
                for room_poly_id in polygons_by_type[dr.DataRoom.data_type]:
                    # check if valid room poly ( just in case that is a room in schedule only >> not placed in model , or unbound, or overlapping with other room)
                    if len(room_polygons[room_poly_id]) > 0:
                        # loop over each room polygon per room...there should only be one...
                        for room_polygon in room_polygons[room_poly_id]:
                            # find overlapping ceiling polygons
                            for ceiling_poly_id in polygons_by_type[
                                dc.DataCeiling.data_type
                            ]:
                                for ceiling_polygon in ceiling_polygons[
                                    ceiling_poly_id
                                ]:
                                    return_value.update(
                                        _intersect_ceiling_vs_room(
                                            ceiling_poly_id,
                                            ceiling_polygon,
                                            room_poly_id,
                                            room_polygon,
                                            data_objects,
                                            level_name,
                                        )
                                    )
                    else:
                        return_value.append_message(
                            "Room with id {} has no valid room poly lines.".format(
                                room_poly_id
                            )
                        )
            else:
                return_value.append_message(
                    "No ceilings found for level: {}".format(
                        data_objects[level_name][0][0].level.name
                    )
                )
        else:
            return_value.append_message(
                "No rooms found for level: {}".format(data_objects[level_name])
            )
    return_value.result = data_objects

    return return_value




from duHast.UI.Objects.WPF.Commands.CommandBase import CommandBase

class RelayCommand(CommandBase):
    def __init__(self, execute):
        """
        A class to handle wpf command bindings.

        Args:
            execute (function): The method to execute when the command is invoked.
        """
        
        super(RelayCommand, self).__init__(execute=execute)
       
    def Execute(self, parameter):
        self.execute(parameter)


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit railing balusters. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from Autodesk.Revit.DB import BuiltInCategory, ElementId, FilteredElementCollector

from duHast.Revit.Common.common import get_ids_from_element_collector
from duHast.Revit.Railings.Utility.merge_lists import merge_into_unique_list
from duHast.Revit.Railings.railings import (
    get_all_railing_type_ids_by_class_and_category,
)


def get_balusters_used_in_pattern(b_pattern):
    """
    Gets list of unique baluster family ids used in a pattern only.
    :param b_pattern: A revit baluster pattern.
    :type b_pattern: Autodesk.Revit.DB.Architecture.BalusterPattern
    :return: List of element ids of baluster family ids.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    for i in range(0, b_pattern.GetBalusterCount()):
        bal_info = b_pattern.GetBaluster(i)
        if (
            bal_info.BalusterFamilyId not in ids
            and bal_info.BalusterFamilyId != ElementId.InvalidElementId
        ):
            ids.append(bal_info.BalusterFamilyId)
    # add excess pattern baluster id
    if (
        b_pattern.ExcessLengthFillBalusterId not in ids
        and b_pattern.ExcessLengthFillBalusterId != ElementId.InvalidElementId
    ):
        ids.append(b_pattern.ExcessLengthFillBalusterId)
    return ids


def get_used_baluster_post_ids(b_post_pattern):
    """
    Gets list of unique baluster posts ids only.
    Includes:
    - CornerPost
    - EndPost
    - StartPost
    :param b_post_pattern: A revit post pattern.
    :type b_post_pattern: Autodesk.Revit.DB.Architecture.PostPattern
    :return: List of element ids of baluster family ids.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    # get corner post
    if b_post_pattern.CornerPost.BalusterFamilyId != ElementId.InvalidElementId:
        ids.append(b_post_pattern.CornerPost.BalusterFamilyId)
    # get end post id
    if (
        b_post_pattern.EndPost.BalusterFamilyId != ElementId.InvalidElementId
        and b_post_pattern.EndPost.BalusterFamilyId not in ids
    ):
        ids.append(b_post_pattern.EndPost.BalusterFamilyId)
    # get start post id
    if (
        b_post_pattern.StartPost.BalusterFamilyId != ElementId.InvalidElementId
        and b_post_pattern.StartPost.BalusterFamilyId not in ids
    ):
        ids.append(b_post_pattern.StartPost.BalusterFamilyId)
    return ids


def get_used_baluster_per_tread(b_placement):
    """
    Gets the id of the baluster per stair tread.
    :param b_placement: A baluster placement element.
    :type b_placement: Autodesk.Revit.DB.Architecture.BalusterPlacement
    :return: List of element ids of baluster family ids.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    # get baluster per tread
    if b_placement.BalusterPerTreadFamilyId != ElementId.InvalidElementId:
        ids.append(b_placement.BalusterPerTreadFamilyId)
    return ids


def get_all_baluster_symbols(doc):
    """
    Gets all baluster symbols (fam types) in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of baluster symbols (types).
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    col = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_StairsRailingBaluster)
        .WhereElementIsElementType()
    )
    return col


def get_all_baluster_symbols_ids(doc):
    """
    Gets all baluster symbol (fam type) ids in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids of baluster family ids.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    col = get_all_baluster_symbols(doc)
    ids = get_ids_from_element_collector(col)
    return ids


def get_baluster_types_from_railings(doc):
    """
    Gets a list of unique baluster symbol (fam type) ids used in railing types in the model.
    Incl:
    - baluster patterns
    - baluster posts
    - baluster per stair 
    There can be additional baluster symbols in the model. Those belong to loaded families which are not used in\
        any railing type definition.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids of baluster family ids.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    ids = []
    railing_type_ids = get_all_railing_type_ids_by_class_and_category(doc)
    for rt_id in railing_type_ids:
        el = doc.GetElement(rt_id)
        # put into try catch since some rail types have no balusters ...top rail
        try:
            baluster_placement = el.BalusterPlacement
            ids_pattern = get_balusters_used_in_pattern(
                baluster_placement.BalusterPattern
            )
            ids_posts = get_used_baluster_post_ids(baluster_placement.PostPattern)
            ids_per_tread = get_used_baluster_per_tread(baluster_placement)
            # build overall ids list
            ids = merge_into_unique_list(ids, ids_pattern)
            ids = merge_into_unique_list(ids, ids_posts)
            ids = merge_into_unique_list(ids, ids_per_tread)
        except:
            pass
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a Revit detail items utility functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


def build_detail_type_ids_dictionary(collector):
    """
    Returns the dictionary keys is autodesk.revit.db element type as string and values are available type ids.

    :param collector: A filtered element collector containing detail component types.
    :type collector: Autodesk.Revit.DB.FilteredElementCollector

    :return: Dictionary where key is the element type as string and value is a list of all type ids belonging to the element type.
    :rtype: dic{str:list[Autodesk.Revit.DB.ElementId]}
    """

    dic = {}
    for c in collector:
        if str(c.GetType()) in dic:
            if c.Id not in dic[str(c.GetType())]:
                dic[str(c.GetType())].append(c.Id)
        else:
            dic[str(c.GetType())] = [c.Id]
    return dic


def build_dependent_elements_dictionary(doc, collector):
    """
    Returns the dictionary keys is autodesk.revit.db element type as string and values are elements of that type.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param collector: A filtered element collector containing elements.
    :type collector: Autodesk.Revit.DB.FilteredElementCollector

    :return: Dictionary where key is the element type as string and value is a list of all elements belonging to the element type.
    :rtype: dic{str:list[Autodesk.Revit.DB.Element]}
    """

    dic = {}
    for c in collector:
        el = doc.GetElement(c)
        if str(el.GetType()) in dic:
            if c not in dic[str(el.GetType())]:
                dic[str(el.GetType())].append(c)
        else:
            dic[str(el.GetType())] = [c]
    return dic



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Base class for objects.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This class provides some utility functions to all child classes:

- __repr__() a way of providing detailed debug output through print
- to_json() A json formatted dump of the class

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json
from collections import OrderedDict
from duHast.Utilities.utility import encode_ascii

"""
The `Base` class is a parent class that provides common functionalities and methods for its subclasses. It includes a constructor, a debug output method, a comparison method, a hash method, a method to convert the instance to JSON, a method to convert string properties to UTF-8 in JSON conversion, a method to check if an object is a Python primitive, and a method to convert the class to a dictionary.

Example Usage:
    # Creating a subclass of Base
    class MyClass(Base):
        def __init__(self, name):
            super(MyClass, self).__init__()
            self.name = name

    # Creating an instance of MyClass
    obj = MyClass("example")

    # Printing the debug output
    print(obj)

    # Comparing two instances
    obj1 = MyClass("example")
    obj2 = MyClass("example")
    print(obj1 == obj2)

    # Converting the instance to JSON
    json_data = obj.to_json()

    # Converting the instance to JSON with UTF-8 string properties
    json_data_utf = obj.to_json_utf()

    # Converting the instance to a dictionary
    dict_data = obj.class_to_dict()

Main functionalities:
- The `Base` class allows for multi-inheritance in its subclasses.
- It provides a debug output method that returns a string representation of the class properties.
- It provides a comparison method that checks if two instances are of the same class.
- It provides a hash method required for the custom comparison method.
- It provides a method to convert the instance to JSON.
- It provides a method to convert string properties to UTF-8 in JSON conversion.
- It provides a method to check if an object is a Python primitive.
- It provides a method to convert the class to a dictionary.

Methods:
- `__init__(self, **kwargs)`: The class constructor that forwards all unused arguments to the super class.
- `__repr__(self)`: Enables detailed debug output of all class properties.
- `__eq__(self, other)`: Custom compare is equal override.
- `__hash__(self)`: Custom hash override.
- `to_json(self)`: Convert the instance of this class to JSON.
- `string_to_utf(self, o)`: Convert any properties stored as string to UTF-8 in JSON conversion.
- `to_json_utf(self)`: Convert the instance of this class to JSON with UTF-8 string properties.
- `_is_primitive(self, obj)`: Checks whether an object is a Python primitive.
- `class_to_dict(self)`: Returns all class properties and their values as a dictionary.

Fields:
- No specific fields are defined in the `Base` class.
"""


class Base(object):
    def __init__(self, **kwargs):
        """
        Class constructor
        """

        # this class inherits from object directly

        super(Base, self).__init__()

    def __repr__(self):
        """
        Enables detailed debug output of all class properties using: rep(obj)

        :return: A string listing class properties and their respective values.
        :rtype: string
        """

        return "{}({})".format(
            self.__class__.__name__,
            ", ".join("{}={!r}".format(k, v) for k, v in self.__dict__.items()),
        )

    def __str__(self):
        return self.default_format()
    
    def default_format(self):
        """
        Provides a formatted output of all class properties and their values without indentation.
        
        :return: A string representing all class properties and their values
        :rtype: str
        """
        output = []
        for attr_name, attr_value in self.__dict__.items():
            if isinstance(attr_value, list):
                output.append("{}:".format(attr_name))
                output.append(self._default_format_list(attr_value))
            elif isinstance(attr_value, dict):
                output.append("{}:".format(attr_name))
                output.append(self._default_format_dict(attr_value))
            elif isinstance(attr_value, Base):
                output.append("{}:".format(attr_name))
                output.append(attr_value.default_format())
            else:
                output.append("{}: {}".format(attr_name, attr_value))
        return "\n".join(output)

    def _default_format_list(self, lst):
        """
        A helper function to format list properties without indentation.
        
        :param lst: A list
        :type lst: []
        :return: A string formatted representation of the list passed in without indentation.
        :rtype: str
        """
        output = []
        for item in lst:
            if isinstance(item, Base):
                output.append(item.default_format())
            elif isinstance(item, dict):
                output.append(self._default_format_dict(item))
            else:
                output.append(str(item))
        return "\n".join(output)

    def _default_format_dict(self, d):
        """
        A helper function to format dictionary properties without indentation.
        
        :param d: A dictionary
        :type d: {}
        :return: A string formatted representation of the dictionary passed in without indentation.
        :rtype: str
        """
        output = []
        for key, value in d.items():
            output.append("{}:".format(key))
            if isinstance(value, Base):
                output.append(value.default_format())
            elif isinstance(value, dict):
                output.append(self._default_format_dict(value))
            elif isinstance(value, list):
                output.append("{}:".format(key))
                output.append(self._default_format_list(value))
            else:
                output.append(str(value))
        return "\n".join(output)

    
    def formatted_indented_str(self, indent=0, indent_character=" "):
        """
        formatted output including indentation

        :param indent: The level of indentation, defaults to 0
        :type indent: int, optional
        :param indent_character: the indentation character used, defaults to " "
        :type indent_character: str

        :return: A string representing all class properties and their values
        :rtype: str
        """

        output = []
        for attr_name, attr_value in self.__dict__.items():
            if isinstance(attr_value, list):
                output.append(indent_character * indent + "{}:".format(attr_name))
                output.append(self._indented_format_list(attr_value, indent + 2, indent_character))
            elif isinstance(attr_value, dict):
                output.append(indent_character * indent + "{}:".format(attr_name))
                output.append(self._indented_format_dict(attr_value, indent + 2, indent_character))
            elif isinstance(attr_value, Base):
                output.append(indent_character * indent + "{}".format(attr_name, indent))
                output.append(attr_value.formatted_indented_str(indent + 2, indent_character))
            else:
                output.append(indent_character * indent + "{}: {}".format(attr_name, attr_value))
        return "\n".join(output)

    def _indented_format_list(self, lst, indent, indent_character):
        """
        A helper function to format list properties

        :param lst: A list
        :type lst: []
        :param indent: level of indentation to be provided to the string output
        :type indent: int

        :return: A string formatted representation of the list past in with indentation.
        :rtype: str
        """
        output = []
        for item in lst:
            if isinstance(item, Base):
                output.append(item.formatted_indented_str(indent, indent_character))
            elif isinstance(item, dict):
                output.append(self._indented_format_dict(item, indent, indent_character))
            else:
                output.append(indent_character * indent + str(item))
        return "\n".join(output)

    def _indented_format_dict(self, d, indent, indent_character):
        """
        A helper function to format dictionary properties.

        :param d: A dictionary
        :type d: {}
        :param indent: level of indentation to be provided to the string output
        :type indent: int

        :return: A string formatted representation of the dictionary past in with indentation.
        :rtype: str
        """
        output = []
        for key, value in d.items():
            output.append(indent_character * indent + "{}:".format(key))
            if isinstance(value, Base):
                output.append(value.formatted_indented_str(indent + 2, indent_character))
            elif isinstance(value, dict):
                output.append(self._indented_format_dict(value, indent + 2, indent_character))
            elif isinstance(value, list):
                output.append(indent_character * (indent + 2) + "{}:".format(key))
                output.append(self._indented_format_list(value, indent + 4, indent_character))
            else:
                output.append(indent_character * (indent + 2) + str(value))
        return "\n".join(output)

    def __eq__(self, other):
        """
        Custom compare is equal override

        :param other: Another instance of this class
        :type other: :class:`.Base`
        :return: True if other is a base instance, otherwise False.
        :rtype: Bool
        """

        return isinstance(other, self.__class__)

    # python 2.7 needs custom implementation of not equal
    def __ne__(self, other):
        return not self.__eq__(other=other)

    def __hash__(self):
        """
        Custom hash override

        Required due to custom __eq__ override present in this class
        """
        try:
            return hash(self.__class__)
        except Exception as e:
            raise ValueError(
                "Exception {} occurred in {} with values: name:{}".format(
                    e, self.data_type, self.name
                )
            )
            

    def to_json(self):
        """
        Convert the instance of this class to JSON, including public attributes and properties.

        :return: A JSON object.
        :rtype: str
        """

        return json.dumps(self.class_to_dict(), ensure_ascii=False)

    def to_json_ordered(self):
        """
        Convert the instance of this class to JSON, including public attributes and properties.
        Properties and attributes are ordered alphabetically.
        Note: This method takes about double the time of to_json()

        :return: A JSON object.
        :rtype: str
        """

        return json.dumps(self.class_to_ordered_dict(), ensure_ascii=False)
    
    def to_json_utf(self):
        """
        Convert the instance of this class to json, any string properties are converted to utf-8

        :return: A Json object.
        :rtype: json
        """

        json_string = self.to_json()
        ascii_encoded = encode_ascii(json_string)
        return ascii_encoded
    
    
    def to_json_utf_ordered(self):
        """
        Convert the instance of this class to json, any string properties are converted to utf-8
        Properties and attributes are ordered alphabetically.
        Note: This method takes about double the time of to_json_utf()
        :return: A Json object.
        :rtype: json
        """

        json_string = self.to_json_ordered()
        ascii_encoded = encode_ascii(json_string)
        return ascii_encoded


    def _is_primitive(self, obj):
        """
        Checks whether object past in is a python primitive

        :param obj: The object to be tested.
        :type obj: obj
        :return: True if object is a python primitive, Otherwise False.
        :rtype: Bool
        """

        return isinstance(obj, (int, float, str, bool))

    def class_to_dict(self):
        """
        Returns all class properties and their values as a dictionary

        :return: A dictionary of all class properties names and their values
        :rtype: {str:var,}
        """

        def serialize(obj):
            """Helper function to recursively serialize objects."""
            if isinstance(obj, Base):
                return obj.class_to_dict()  # Call if it's an instance of Base
            elif isinstance(obj, list):
                return [
                    serialize(item) for item in obj
                ]  # Recursively serialize list items
            elif isinstance(obj, dict):
                return {
                    key: serialize(value) for key, value in obj.items()
                }  # Recursively serialize dict values
            else:
                return obj  # Return the object as is

        if isinstance(self, object):
            class_dict = {}
            # Include instance attributes
            for key, value in self.__dict__.items():
                if not key.startswith("_"):  # Exclude private properties
                    if self._is_primitive(value):
                        class_dict[key] = value
                    else:
                        class_dict[key] = serialize(value)
            # Include properties
            for key in dir(self):
                if isinstance(getattr(type(self), key, None), property):
                    value = getattr(self, key)
                    if self._is_primitive(value):
                        class_dict[key] = value
                    else:
                        class_dict[key] = serialize(value)
            return class_dict
        else:
            return self
    
    def class_to_ordered_dict(self):
        """
        Returns all class properties and their values as an OrderedDict in alphabetical order of the property names.
        (This takes about double the time of class_to_dict)

        :return: An OrderedDict of all class properties names and their values
        :rtype: OrderedDict
        """

        def serialize(obj):
            """Helper function to recursively serialize objects."""
            if isinstance(obj, Base):
                return obj.class_to_ordered_dict()  # Call if it's an instance of Base
            elif isinstance(obj, list):
                return [
                    serialize(item) for item in obj
                ]  # Recursively serialize list items
            elif isinstance(obj, dict):
                return OrderedDict(
                sorted(
                    ((key, serialize(value)) for key, value in obj.items()),
                    key=lambda item: item[0]
                )
            )  # Recursively serialize and sort dict values
            else:
                return obj  # Return the object as is

        if isinstance(self, object):
            items = []
            class_dict = OrderedDict()
            # Include instance attributes
            for key, value in self.__dict__.items():
                if not key.startswith("_"):  # Exclude private properties
                    if self._is_primitive(value):
                        items.append((key, value))
                    else:
                        items.append((key, serialize(value)))
            # Include properties
            for key in dir(self):
                if isinstance(getattr(type(self), key, None), property):
                    value = getattr(self, key)
                    if self._is_primitive(value):
                        items.append((key, value))
                    else:
                        items.append((key, serialize(value)))
            
            # Sort items alphabetically by key
            items.sort(key=lambda item: item[0])
            
            # Add sorted items to OrderedDict
            for key, value in items:
                class_dict[key] = value
            
            return class_dict
        else:
            return self

